from collections.abc import Callable
from enum import (
    Enum,
)
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

from networkx import DiGraph

from blends.syntax.node_attributes import VALID_SYNTAX_ATTRIBUTES


class Language(Enum):
    CSHARP = "c_sharp"
    DART = "dart"
    GO = "go"
    HCL = "hcl"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    JSON = "json"
    KOTLIN = "kotlin"
    NOT_SUPPORTED = "not_supported"
    PHP = "php"
    PYTHON = "python"
    RUBY = "ruby"
    SCALA = "scala"
    SWIFT = "swift"
    TYPESCRIPT = "tsx"
    YAML = "yaml"


LANGUAGE_TO_EXTENSIONS: dict[Language, tuple[str, ...]] = {
    Language.CSHARP: (".cs",),
    Language.DART: (".dart",),
    Language.GO: (".go",),
    Language.HCL: (".hcl", ".tf"),
    Language.JAVA: (".java",),
    Language.JAVASCRIPT: (".js", ".jsx"),
    Language.JSON: (".json",),
    Language.KOTLIN: (".kt", ".ktm", ".kts"),
    Language.PHP: (".php",),
    Language.PYTHON: (".py",),
    Language.RUBY: (".rb",),
    Language.SCALA: (".scala",),
    Language.SWIFT: (".swift",),
    Language.TYPESCRIPT: (".ts", ".tsx"),
    Language.YAML: (".yaml", ".yml"),
}

EXTENSION_TO_LANGUAGE: dict[str, Language] = {
    ext: lang for lang, exts in LANGUAGE_TO_EXTENSIONS.items() for ext in exts
}


def decide_language(path: str) -> Language:
    file_extension = "." + path.split(".")[-1]
    return EXTENSION_TO_LANGUAGE.get(file_extension, Language.NOT_SUPPORTED)


SUPPORTED_MULTIFILE = {
    Language.JAVA,
}


class Content(NamedTuple):
    content_bytes: bytes
    content_str: str
    language: Language
    path: Path


class CustomParser(NamedTuple):
    validator: Callable[[Content], bool]
    parser: Callable[[Content], Content | None]


NId = str

if TYPE_CHECKING:

    class BaseGraph(DiGraph[NId]): ...
else:

    class BaseGraph(DiGraph): ...


class Graph(BaseGraph):
    def __init__(self) -> None:  # noqa: D107
        super().__init__()
        self.nodes_by_type: dict[str, list[NId]] = {}

    def add_node(self, node_for_adding: NId, **attr: object) -> None:
        super().add_node(node_for_adding, **attr)
        if label_type := attr.get("label_type"):
            self.nodes_by_type.setdefault(str(label_type), []).append(node_for_adding)

    def update_node(self, node_id: NId, attributes: dict[str, object]) -> None:
        self.nodes[node_id].update(attributes)


class SyntaxGraph(Graph):
    def add_node(self, node_for_adding: NId, **attr: object) -> None:
        invalid_keys = attr.keys() - VALID_SYNTAX_ATTRIBUTES
        if invalid_keys:
            msg = f"Invalid node attribute keys: {invalid_keys!r}"
            raise ValueError(msg)
        super().add_node(node_for_adding, **attr)

    def update_node(self, node_id: NId, attributes: dict[str, object]) -> None:
        invalid_keys = attributes.keys() - VALID_SYNTAX_ATTRIBUTES
        if invalid_keys:
            msg = f"Invalid node attribute keys: {invalid_keys!r}"
            raise ValueError(msg)
        super().update_node(node_id, attributes)

    def update_node_attribute(self, node_id: NId, key: str, value: object) -> None:
        if key not in VALID_SYNTAX_ATTRIBUTES:
            msg = f"Invalid node attribute key: {key!r}"
            raise ValueError(msg)
        self.nodes[node_id][key] = value


class GraphPair(NamedTuple):
    ast_graph: Graph | None
    syntax_graph: SyntaxGraph | None


class GraphShard(NamedTuple):
    path: str
    graph: Graph
    syntax_graph: SyntaxGraph
    content_as_str: str
    file_size: int


class GraphDB(NamedTuple):
    context: dict[Language, dict[str, dict]]  # type: ignore[type-arg]
    shards: dict[str, GraphShard]
    properties: dict[str, dict[str, str]] = {}  # noqa: RUF012

    def get_path_shard(self, path: str) -> GraphShard | None:
        return self.shards.get(path)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.modifiers import (
    build_modifiers_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    c_ids = match_ast_group_d(args.ast_graph, args.n_id, "attribute")
    return build_modifiers_node(args, iter(c_ids))



from blends.models import (
    NId,
)
from blends.syntax.builders.named_argument import (
    build_named_argument_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    field_node = args.ast_graph.nodes[args.n_id]["label_field_name"]
    return build_named_argument_node(args, None, field_node)



from blends.symbolic_evaluation.models import (
    Evaluator,
    MissingSymbolicEvalError,
    SymbolicEvalArgs,
    SymbolicEvaluation,
)

EVALUATORS: dict[str, Evaluator] = {}


def register_evaluators(evaluators: dict[str, Evaluator]) -> None:
    EVALUATORS.update(evaluators)


def generic(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    node_type = args.graph.nodes[args.n_id]["label_type"]
    evaluator = EVALUATORS.get(node_type)
    if not evaluator:
        exc_log = f"Missing symbolic evaluator {node_type}"
        raise MissingSymbolicEvalError(exc_log)

    if args.n_id in args.evaluation:
        return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)

    return evaluator(args)



from collections.abc import (
    Iterator,
)

from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_string_literal_node(
    args: SyntaxGraphArgs,
    value: str,
    c_ids: Iterator[NId] | None = None,
) -> NId:
    if value.startswith(("'", '"', "`")):
        value = value[1:-1]

    args.syntax_graph.add_node(
        args.n_id,
        value=value,
        value_type="string",
        label_type="Literal",
    )

    if c_ids:
        for c_id in c_ids:
            args.syntax_graph.add_edge(
                args.n_id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    return args.n_id



from blends.models import (
    NId,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    attrs = args.ast_graph.nodes[args.n_id]

    var_id = attrs["label_field_name"]
    var_name = node_to_str(args.ast_graph, var_id)

    type_id = attrs.get("label_field_type")
    var_type = node_to_str(graph, type_id) if type_id else None

    return build_parameter_node(args=args, variable=var_name, variable_type=var_type, value_id=None)



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_PRECEDENCE,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.stack.validation import (
    validate_stack_graph_graph,
    validate_stack_graph_node,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_push_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_pop_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "bar")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 5)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_DEFINITION, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_push_scoped_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.add_node("scope_1")
    graph.update_node_attribute(
        "node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "baz")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, "scope_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_scope_node_exported() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_scope_node_not_exported() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_root_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.ROOT.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_drop_scopes_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_jump_to_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_valid_pop_scoped_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute(
        "node_1", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "qux")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, 0)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_DEFINITION, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_missing_symbol_on_push_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Missing symbol" in errors[0]
    assert "node_1" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_empty_symbol_on_push_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Missing symbol" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_missing_scope_on_push_scoped_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute(
        "node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Missing scope" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_missing_is_exported_on_scope_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, None)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Missing is_exported" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_symbol_on_scope_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "unexpected")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=False)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected symbol" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_scope_on_push_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, "scope_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected scope" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_is_exported_on_push_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected is_exported" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_missing_symbol_on_pop_scoped_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute(
        "node_1", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_DEFINITION, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Missing symbol" in errors[0]
    assert "node_1" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_symbol_on_drop_scopes_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "unexpected")

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected symbol" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_symbol_on_jump_to_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "unexpected")

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected symbol" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_scope_on_drop_scopes_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, "scope_1")

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected scope" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_scope_on_jump_to_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, "scope_1")

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected scope" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_is_exported_on_drop_scopes_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected is_exported" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_is_exported_on_jump_to_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected is_exported" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unexpected_is_exported_on_pop_scoped_symbol_node() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute(
        "node_1", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_DEFINITION, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unexpected is_exported" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_invalid_precedence() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_PRECEDENCE, "not_an_int")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Invalid precedence" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_unknown_kind() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, "UnknownNodeKind")

    errors = validate_stack_graph_node(graph, "node_1")

    assert len(errors) == 1
    assert "Unknown stack graph kind" in errors[0]
    assert "UnknownNodeKind" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_returns_error_list() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_1", STACK_GRAPH_SCOPE, "unexpected")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_EXPORTED, value=True)
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_node(graph, "node_1")

    assert isinstance(errors, list)
    assert len(errors) == 3  # noqa: PLR2004
    assert all(isinstance(error, str) for error in errors)


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_ignores_nodes_without_kind() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1", label_type="SomeNode")

    errors = validate_stack_graph_node(graph, "node_1")

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_stack_graph_graph_all_nodes() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)
    graph.add_node("node_2")
    graph.update_node_attribute("node_2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("node_2", STACK_GRAPH_SYMBOL, None)
    graph.update_node_attribute("node_2", STACK_GRAPH_IS_DEFINITION, value=True)
    graph.add_node("node_3")

    errors = validate_stack_graph_graph(graph)

    assert len(errors) == 1
    assert "Missing symbol" in errors[0]
    assert "node_2" in errors[0]


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_stack_graph_graph_specific_nodes() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("node_1", STACK_GRAPH_SYMBOL, "foo")
    graph.update_node_attribute("node_1", STACK_GRAPH_IS_REFERENCE, value=True)
    graph.add_node("node_2")
    graph.update_node_attribute("node_2", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("node_2", STACK_GRAPH_SYMBOL, None)

    errors = validate_stack_graph_graph(graph, node_ids=["node_1"])

    assert errors == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_validate_stack_graph_graph_with_new_node_kinds() -> None:
    graph = SyntaxGraph()
    graph.add_node("node_1")
    graph.update_node_attribute("node_1", STACK_GRAPH_KIND, StackGraphNodeKind.DROP_SCOPES.value)
    graph.add_node("node_2")
    graph.update_node_attribute("node_2", STACK_GRAPH_KIND, StackGraphNodeKind.JUMP_TO.value)
    graph.add_node("node_3")
    graph.update_node_attribute(
        "node_3", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_3", STACK_GRAPH_SYMBOL, "bar")
    graph.update_node_attribute("node_3", STACK_GRAPH_IS_DEFINITION, value=True)
    graph.add_node("node_4")
    graph.add_node("scope_1")
    graph.update_node_attribute(
        "node_4", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value
    )
    graph.update_node_attribute("node_4", STACK_GRAPH_SYMBOL, "baz")
    graph.update_node_attribute("node_4", STACK_GRAPH_SCOPE, "scope_1")
    graph.update_node_attribute("node_4", STACK_GRAPH_IS_REFERENCE, value=True)

    errors = validate_stack_graph_graph(graph)

    assert errors == []



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    text = node_to_str(graph, args.n_id)
    if text.startswith(('f"', "f'")):
        text = text.lstrip("f")
        ints = match_ast_group_d(graph, args.n_id, "interpolation")
        int_subs = [graph.nodes[_id]["label_field_expression"] for _id in ints]
        if len(int_subs) > 0:
            return build_string_literal_node(args, text, iter(int_subs))

    return build_string_literal_node(args, text)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.array_node import (
    build_array_node,
)
from blends.syntax.builders.literal import (
    build_literal_node,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    child_id = adj_ast(args.ast_graph, args.n_id)[0]
    if args.ast_graph.nodes[child_id]["label_type"] == "plain_scalar":
        lit_id = adj_ast(args.ast_graph, child_id)[0]
        literal_type = args.ast_graph.nodes[lit_id]["label_type"]
        literal_text = args.ast_graph.nodes[lit_id]["label_text"]
        if literal_type in {"integer_scalar", "float_scalar"}:
            return build_literal_node(args, literal_text, "number")
        if literal_type == "boolean_scalar":
            return build_literal_node(args, literal_text, "bool")
        if literal_type == "string_scalar":
            return build_string_literal_node(args, literal_text)

    if args.ast_graph.nodes[child_id]["label_type"] == "flow_sequence":
        valid_childs = [
            child
            for child in adj_ast(args.ast_graph, child_id)
            if args.ast_graph.nodes[child]["label_type"] not in ("[", "]", "{", "}", ",", ";")
        ]
        return build_array_node(args, valid_childs)

    text_node = node_to_str(args.ast_graph, args.n_id)
    return build_string_literal_node(args, text_node)



from dataclasses import dataclass

from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path.errors import (
    _raise_empty_symbol_stack,
    _raise_incorrect_popped_symbol,
    _raise_missing_attached_scope_list,
    _raise_unexpected_attached_scope_list,
)
from blends.stack.partial_path.partial_stacks import (
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.view import StackGraphView


@dataclass(frozen=True, slots=True)
class HalfOpenResult:
    symbol_stack: PartialSymbolStack
    scope_stack: PartialScopeStack


def _pop_symbol(symbol_stack: PartialSymbolStack) -> tuple[PartialScopedSymbol, PartialSymbolStack]:
    if not symbol_stack.symbols:
        _raise_empty_symbol_stack()
    head = symbol_stack.symbols[0]
    tail = PartialSymbolStack(symbols=symbol_stack.symbols[1:], variable=symbol_stack.variable)
    return head, tail


def halfopen_closed_partial_precondition(
    view: StackGraphView,
    join_node_index: int,
    symbol_stack_precondition: PartialSymbolStack,
    scope_stack_precondition: PartialScopeStack,
) -> HalfOpenResult:
    kind = view.kind_at(join_node_index)
    if kind not in {
        StackGraphNodeKind.POP_SYMBOL.value,
        StackGraphNodeKind.POP_SCOPED_SYMBOL.value,
    }:
        return HalfOpenResult(
            symbol_stack=symbol_stack_precondition,
            scope_stack=scope_stack_precondition,
        )
    popped_symbol, updated_symbol_stack = _pop_symbol(symbol_stack_precondition)
    expected_symbol_id = view.symbol_id_at(join_node_index)
    if expected_symbol_id is None or popped_symbol.symbol_id != expected_symbol_id:
        _raise_incorrect_popped_symbol()
    if kind == StackGraphNodeKind.POP_SYMBOL.value:
        if popped_symbol.scopes is not None:
            _raise_unexpected_attached_scope_list()
        return HalfOpenResult(
            symbol_stack=updated_symbol_stack,
            scope_stack=scope_stack_precondition,
        )
    if popped_symbol.scopes is None:
        _raise_missing_attached_scope_list()
    return HalfOpenResult(
        symbol_stack=updated_symbol_stack,
        scope_stack=popped_symbol.scopes,
    )


def halfopen_closed_partial_postcondition(
    view: StackGraphView,
    join_node_index: int,
    symbol_stack_postcondition: PartialSymbolStack,
    scope_stack_postcondition: PartialScopeStack,
) -> HalfOpenResult:
    kind = view.kind_at(join_node_index)
    if kind not in {
        StackGraphNodeKind.PUSH_SYMBOL.value,
        StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value,
    }:
        return HalfOpenResult(
            symbol_stack=symbol_stack_postcondition,
            scope_stack=scope_stack_postcondition,
        )
    popped_symbol, updated_symbol_stack = _pop_symbol(symbol_stack_postcondition)
    expected_symbol_id = view.symbol_id_at(join_node_index)
    if expected_symbol_id is None or popped_symbol.symbol_id != expected_symbol_id:
        _raise_incorrect_popped_symbol()
    if kind == StackGraphNodeKind.PUSH_SYMBOL.value:
        if popped_symbol.scopes is not None:
            _raise_unexpected_attached_scope_list()
        return HalfOpenResult(
            symbol_stack=updated_symbol_stack,
            scope_stack=scope_stack_postcondition,
        )
    if popped_symbol.scopes is None:
        _raise_missing_attached_scope_list()
    return HalfOpenResult(
        symbol_stack=updated_symbol_stack,
        scope_stack=scope_stack_postcondition,
    )



from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from blends.stack.partial_path.partial_stacks import (
        PartialScopeStack,
        PartialSymbolStack,
    )
    from blends.stack.partial_path.variables import (
        ScopeStackVariable,
        SymbolStackVariable,
    )


@dataclass(frozen=True, slots=True)
class _SymbolUnifyBindings:
    symbol_bindings: "PartialSymbolStackBindings"
    scope_bindings: "PartialScopeStackBindings"


@dataclass(slots=True)
class PartialScopeStackBindings:
    bindings: dict["ScopeStackVariable", "PartialScopeStack"]

    @classmethod
    def new(cls) -> "PartialScopeStackBindings":
        return cls(bindings={})

    def get(self, variable: "ScopeStackVariable") -> "PartialScopeStack | None":
        return self.bindings.get(variable)

    def add(self, variable: "ScopeStackVariable", scopes: "PartialScopeStack") -> None:
        old = self.bindings.get(variable)
        if old is None:
            self.bindings[variable] = scopes
            return
        unified = scopes.unify(old, self)
        self.bindings[variable] = unified


@dataclass(slots=True)
class PartialSymbolStackBindings:
    bindings: dict["SymbolStackVariable", "PartialSymbolStack"]

    @classmethod
    def new(cls) -> "PartialSymbolStackBindings":
        return cls(bindings={})

    def get(self, variable: "SymbolStackVariable") -> "PartialSymbolStack | None":
        return self.bindings.get(variable)

    def add(
        self,
        variable: "SymbolStackVariable",
        symbols: "PartialSymbolStack",
        scope_bindings: "PartialScopeStackBindings",
    ) -> None:
        old = self.bindings.get(variable)
        if old is None:
            self.bindings[variable] = symbols
            return
        unified = symbols.unify(old, self, scope_bindings)
        self.bindings[variable] = unified



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.switch_statement import (
    build_switch_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    body_id = graph.nodes[args.n_id]["label_field_body"]
    value_id = graph.nodes[args.n_id]["label_field_condition"]
    if graph.nodes[value_id]["label_type"] == "parenthesized_expression":
        value_id = match_ast(graph, value_id).get("__1__")
    return build_switch_statement_node(args, body_id, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    match_ast,
)
from blends.syntax.builders.switch_statement import (
    build_switch_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    if (
        (value_id := args.ast_graph.nodes[args.n_id]["label_field_condition"])
        and (graph.nodes[value_id].get("label_type") == "parenthesized_expression")
        and (c_ids := match_ast(graph, value_id))
        and (clean_val := c_ids.get("__1__"))
    ):
        value_id = clean_val

    body_id = args.ast_graph.nodes[args.n_id]["label_field_body"]
    return build_switch_statement_node(args, body_id, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    possible_operands = {
        "member_access_expression",
        "variable_name",
        "scoped_property_access_expression",
        "subscript_expression",
    }

    operand_n_id = None
    operator_n_id = None
    childs = adj_ast(graph, args.n_id)

    for n_id in childs:
        if graph.nodes[n_id].get("label_type") in possible_operands:
            operand_n_id = n_id
        else:
            operator_n_id = n_id

    if operand_n_id and operator_n_id:
        operator = graph.nodes[operator_n_id]["label_text"]
    else:
        operand_n_id = childs[0]
        operator = childs[-1]

    return build_unary_expression_node(args, operator, operand_n_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    decl_id = match_ast_d(graph, args.n_id, "variable_declaration")
    if not decl_id:
        decl_id = adj_ast(graph, args.n_id)[0]

    var_type_id = graph.nodes[decl_id]["label_field_type"]
    var_type = node_to_str(graph, var_type_id)
    value_id = None

    var_decl_id = match_ast_d(graph, decl_id, "variable_declarator")
    if not var_decl_id:
        var_decl_id = adj_ast(graph, decl_id)[-1]

    if identifier_ids := match_ast_group_d(graph, var_decl_id, "identifier"):
        var_name = node_to_str(graph, identifier_ids[0])
    else:
        var_name = node_to_str(graph, adj_ast(graph, var_decl_id)[0])

    value_id = None
    if len(identifier_ids) > 1:
        value_id = identifier_ids[-1]
    elif (declarator_id := match_ast_d(graph, decl_id, "variable_declarator")) and (
        declarator_childs := list(
            filter(
                lambda child: graph.nodes[child]["label_type"] not in ("identifier", "="),
                list(adj_ast(graph, declarator_id)),
            ),
        )
    ):
        value_id = declarator_childs[0]

    return build_variable_declaration_node(args, var_name, var_type, value_id)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxCfgArgs,
)


def build(args: SyntaxCfgArgs) -> NId:
    key_id = args.graph.nodes[args.n_id].get("key_id")
    value_id = args.graph.nodes[args.n_id].get("value_id")
    if key_id and value_id:
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(key_id, args.nxt_id)),
            label_cfg="CFG",
        )
        args.graph.add_edge(
            args.n_id,
            args.generic(args.fork(value_id, args.nxt_id)),
            label_cfg="CFG",
        )
    return args.n_id



from blends.models import (
    NId,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    attrs = args.ast_graph.nodes[args.n_id]
    var_id = attrs["label_field_name"]
    if graph.nodes[var_id]["label_type"] == "pattern" and (
        identifier := graph.nodes[var_id].get("label_field_bound_identifier")
    ):
        var_id = node_to_str(graph, identifier)
    val_id = attrs.get("label_field_value")
    if not val_id:
        val_id = attrs.get("label_field_computed_value")
    return build_variable_declaration_node(args, var_id, None, val_id)



from blends.models import NId
from blends.syntax.models import ReaderLogicError


def check_reader_element_consistency(*args: NId | None) -> None:
    if any(arg is None for arg in args):
        msg = "Inconsistent reader element"
        raise ReaderLogicError(msg)


def ensure_nid(value: NId | None) -> NId:
    if value is None:
        msg = "Inconsistent reader element, expected NId but got None"
        raise ReaderLogicError(msg)
    return value



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.object import (
    build_object_node,
)
from blends.syntax.builders.object_creation import (
    build_object_creation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    valid_parameters = {
        "literal_value",
        "qualified_type",
    }
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    expected_children_for_composite_literal = 2
    if (
        len(c_ids) == expected_children_for_composite_literal
        and graph.nodes[c_ids[1]]["label_type"] == "literal_value"
        and len(match_ast_group_d(graph, c_ids[1], "keyed_element")) > 0
    ):
        name = node_to_str(graph, c_ids[0])
        if graph.nodes[c_ids[0]]["label_type"] == "map_type":
            return build_object_creation_node(
                args,
                name=name,
                arguments_id=c_ids[1],
                initializer_id=None,
            )
        return build_object_creation_node(
            args,
            name=name,
            arguments_id=c_ids[1],
            initializer_id=c_ids[0],
        )

    return build_object_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] in valid_parameters),
    )



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    _, *c_ids, _ = adj_ast(graph, args.n_id)

    arg_ids = []

    for _id in c_ids:
        if graph.nodes[_id].get("label_type") == "value_argument":
            if not (graph.nodes[_id].get("label_field_name")) and (
                value_id := graph.nodes[_id].get("label_field_value")
            ):
                arg_ids.append(value_id)
            else:
                arg_ids.append(_id)

    return build_argument_list_node(
        args,
        iter(arg_ids),
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    as_attrs = args.ast_graph.nodes[args.n_id]
    var_id = as_attrs["label_field_left"]
    val_id = as_attrs["label_field_right"]
    op_id = as_attrs["label_field_operator"]
    operation = args.ast_graph.nodes[op_id]["label_text"]
    return build_assignment_node(args, var_id, val_id, operation)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    as_attrs = graph.nodes[args.n_id]

    var_id = as_attrs["label_field_left"]
    if (
        graph.nodes[var_id]["label_type"] == "expression_list"
        and (childs := adj_ast(graph, var_id))
        and len(childs) == 1
    ):
        var_id = childs[0]

    val_id = as_attrs["label_field_right"]
    if (
        graph.nodes[val_id]["label_type"] == "expression_list"
        and (childs := adj_ast(graph, val_id))
        and len(childs) == 1
    ):
        val_id = childs[0]

    op_name = None
    if op_id := as_attrs.get("label_field_operator"):
        op_name = node_to_str(graph, op_id)

    return build_assignment_node(args, var_id, val_id, op_name)



from unittest.mock import patch

import pytest

from blends.models import Graph, Language, NId, SyntaxGraph
from blends.stack.attributes import STACK_GRAPH_KIND, STACK_GRAPH_SYMBOL
from blends.stack.node_kinds import StackGraphNodeKind
from blends.syntax.builders.file import build_file_node
from blends.syntax.builders.module_path import compute_module_path_segments
from blends.syntax.models import SyntaxGraphArgs, SyntaxMetadata


def _make_file_args(*, path: str, next_synthetic_id: int = 100) -> SyntaxGraphArgs:
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": ["1"],
        "next_synthetic_id": next_synthetic_id,
    }

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path=path,
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="1",
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_path_simple_module() -> None:
    assert compute_module_path_segments("main.py") == ["main"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_path_extracts_stem_from_path_with_directories() -> None:
    assert compute_module_path_segments("pkg/mod.py") == ["mod"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_path_extracts_stem_from_absolute_path() -> None:
    assert compute_module_path_segments("/home/user/project/pkg/mod.py") == ["mod"]


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_path_init_file_returns_empty() -> None:
    assert compute_module_path_segments("__init__.py") == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_path_init_in_package_returns_empty() -> None:
    assert compute_module_path_segments("pkg/__init__.py") == []


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_node_adds_root_node() -> None:
    args = _make_file_args(path="pkg/mod.py")
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_file_node(args, c_ids=[])

    root_nid = args.metadata.get("root_nid")
    assert root_nid is not None
    assert root_nid in args.syntax_graph.nodes
    assert args.syntax_graph.nodes[root_nid][STACK_GRAPH_KIND] == StackGraphNodeKind.ROOT.value


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_node_chain_connects_root_to_file_scope() -> None:
    args = _make_file_args(path="pkg/mod.py")
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_file_node(args, c_ids=[])

    graph = args.syntax_graph
    root_nid = args.metadata["root_nid"]
    root_successors = list(graph.successors(root_nid))
    assert len(root_successors) == 1
    mod_node = root_successors[0]
    assert graph.nodes[mod_node][STACK_GRAPH_SYMBOL] == "mod"
    assert graph.nodes[mod_node][STACK_GRAPH_KIND] == StackGraphNodeKind.POP_SYMBOL.value

    mod_successors = list(graph.successors(mod_node))
    assert "1" in mod_successors


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_node_absolute_path_uses_stem_only() -> None:
    args = _make_file_args(path="/home/user/project/pkg/mod.py")
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_file_node(args, c_ids=[])

    graph = args.syntax_graph
    root_nid = args.metadata["root_nid"]
    root_successors = list(graph.successors(root_nid))
    assert len(root_successors) == 1
    mod_node = root_successors[0]
    assert graph.nodes[mod_node][STACK_GRAPH_SYMBOL] == "mod"


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_node_no_module_path_when_stack_graph_disabled() -> None:
    args = _make_file_args(path="pkg/mod.py")
    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_file_node(args, c_ids=[])

    assert args.metadata.get("root_nid") is None


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_node_init_file_does_not_create_root() -> None:
    args = _make_file_args(path="pkg/__init__.py")
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_file_node(args, c_ids=[])

    assert args.metadata.get("root_nid") is None



from blends.models import (
    NId,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    operator_id = graph.nodes[args.n_id]["label_field_operator"]
    operator = graph.nodes[operator_id]["label_text"]
    operand_id = graph.nodes[args.n_id]["label_field_argument"]
    return build_unary_expression_node(args, operator, operand_id)



from __future__ import annotations

from typing import TYPE_CHECKING

from blends.models import Language
from blends.stack.policies.default import DefaultExportPolicy
from blends.stack.policies.go import GoExportPolicy
from blends.stack.policies.java import JavaExportPolicy
from blends.stack.policies.javascript import JavaScriptExportPolicy
from blends.stack.policies.kotlin import KotlinExportPolicy
from blends.stack.policies.python import PythonExportPolicy

if TYPE_CHECKING:
    from blends.stack.policies.base import ExportPolicy

_POLICY_REGISTRY: dict[Language, ExportPolicy] = {
    Language.PYTHON: PythonExportPolicy(),
    Language.JAVA: JavaExportPolicy(),
    Language.GO: GoExportPolicy(),
    Language.JAVASCRIPT: JavaScriptExportPolicy(),
    Language.TYPESCRIPT: JavaScriptExportPolicy(),
    Language.KOTLIN: KotlinExportPolicy(),
    Language.SCALA: KotlinExportPolicy(),
    Language.CSHARP: JavaExportPolicy(),
}

_DEFAULT_POLICY = DefaultExportPolicy()


def get_export_policy(language: Language) -> ExportPolicy:
    return _POLICY_REGISTRY.get(language, _DEFAULT_POLICY)



from blends.models import (
    NId,
)
from blends.query import match_ast_d, match_ast_group_d, pred_ast
from blends.syntax.builders.string_literal import build_string_literal_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import node_to_str


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    text = node_to_str(graph, args.n_id)
    parent = pred_ast(graph, args.n_id)[0]
    c_ids = []
    if graph.nodes[parent]["label_type"] == "subshell" and (
        interpolation := match_ast_d(graph, parent, "interpolation")
    ):
        identifiers = match_ast_group_d(graph, interpolation, "identifier")
        element_refs = match_ast_group_d(graph, interpolation, "element_reference")
        c_ids = identifiers + element_refs
    return build_string_literal_node(args, text, iter(c_ids) if c_ids else None)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    block = graph.nodes[args.n_id].get("label_field_body")
    if not block:
        block = adj_ast(graph, args.n_id)[-1]
    conditional_node = graph.nodes[args.n_id]["label_field_condition"]
    return build_while_statement_node(args, block, conditional_node)



from blends.models import (
    NId,
)
from blends.syntax.builders.assignment import (
    build_assignment_node,
)
from blends.syntax.builders.variable_declaration import (
    build_variable_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    variable_id = n_attrs["label_field_left"]
    value_id = n_attrs["label_field_right"]

    ruby_variable_nodes = {
        "identifier",
        "constant",
        "global_variable",
        "class_variable",
        "instance_variable",
        "exception_variable",
    }
    if graph.nodes[variable_id]["label_type"] in ruby_variable_nodes:
        var_name = node_to_str(graph, variable_id)
        return build_variable_declaration_node(args, var_name, None, value_id)

    return build_assignment_node(args, variable_id, value_id, None)



from blends.models import (
    NId,
)
from blends.syntax.builders.namespace import (
    build_namespace_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    namespace = args.ast_graph.nodes[args.n_id]
    block_id = namespace["label_field_body"]
    name = node_to_str(args.ast_graph, namespace["label_field_name"])
    return build_namespace_node(args, name, block_id)



from collections.abc import (
    Iterator,
)

from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    pop_scoped_symbol_node_attributes,
    pop_symbol_node_attributes,
)
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
    get_next_synthetic_node_id,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _resolve_class_scope_for_parameter(
    variable: str,
    annotation_class_names: set[str] | None,
    class_member_scope_stack: list[tuple[NId, str]],
    *,
    is_classmethod: bool = False,
) -> NId | None:
    if not class_member_scope_stack:
        return None
    if variable == "self":
        return class_member_scope_stack[-1][0]
    if variable == "cls" and is_classmethod:
        return class_member_scope_stack[-1][0]
    if annotation_class_names:
        for scope_nid, class_name in reversed(class_member_scope_stack):
            if class_name in annotation_class_names:
                return scope_nid
    return None


def _setup_stack_graph_parameter(
    args: SyntaxGraphArgs,
    node_id: NId,
    variable: str,
    annotation_class_names: set[str] | None = None,
) -> None:
    if node_id not in args.syntax_graph.nodes:
        return

    precedence = get_next_binding_precedence(args)
    args.syntax_graph.update_node(
        node_id,
        pop_scoped_symbol_node_attributes(
            symbol=variable,
            precedence=precedence,
        ),
    )
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack and (parent_scope := scope_stack[-1]):
        add_edge(
            args.syntax_graph,
            Edge(source=parent_scope, sink=node_id, precedence=precedence),
        )
    class_member_scope_stack = args.metadata.get("class_member_scope_stack", [])
    is_classmethod = args.metadata.get("is_classmethod", False)
    target_scope_nid = _resolve_class_scope_for_parameter(
        variable, annotation_class_names, class_member_scope_stack, is_classmethod=is_classmethod
    )
    if target_scope_nid is not None:
        dot_gateway_id = get_next_synthetic_node_id(args)
        args.syntax_graph.add_node(
            dot_gateway_id,
            label_type="SyntheticDotGateway",
            **pop_symbol_node_attributes(symbol=".", is_definition=False),
        )
        add_edge(args.syntax_graph, Edge(source=node_id, sink=dot_gateway_id))
        add_edge(args.syntax_graph, Edge(source=dot_gateway_id, sink=target_scope_nid))


def build_parameter_node(  # noqa: PLR0913
    *,
    args: SyntaxGraphArgs,
    variable: str | None,
    variable_type: str | None,
    value_id: NId | None,
    c_ids: Iterator[NId] | list[NId] | None = None,
    variable_id: NId | None = None,
    modifier: NId | None = None,
    annotation_class_names: set[str] | None = None,
) -> NId:
    _id = variable_id if variable_id else args.n_id
    args.syntax_graph.add_node(
        _id,
        label_type="Parameter",
    )

    if variable:
        args.syntax_graph.update_node_attribute(_id, "variable", variable)

    if variable_type:
        args.syntax_graph.update_node_attribute(_id, "variable_type", variable_type)

    if value_id:
        args.syntax_graph.update_node_attribute(_id, "value_id", value_id)
        args.syntax_graph.add_edge(
            _id,
            args.generic(args.fork_n_id(value_id)),
            label_ast="AST",
        )

    if c_ids:
        for c_id in c_ids:
            args.syntax_graph.add_edge(
                _id,
                args.generic(args.fork_n_id(c_id)),
                label_ast="AST",
            )

    if modifier:
        args.syntax_graph.update_node_attribute(_id, "parameter_mode", modifier)

    if ctx.has_feature_flag("StackGraph") and variable:
        _setup_stack_graph_parameter(
            args, _id, variable, annotation_class_names=annotation_class_names
        )

    return _id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.class_body import (
    build_class_body_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    _, *c_ids, _ = adj_ast(args.ast_graph, args.n_id)  # do not consider { }
    return build_class_body_node(args, c_ids)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = match_ast_group_d(graph, args.n_id, "value_argument")
    arg_ids = []
    for _id in c_ids:
        val_id = adj_ast(graph, _id)[0]
        if val_id:
            arg_ids.append(val_id)
    return build_argument_list_node(args, iter(arg_ids))



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_unary_expression_node(
    args: SyntaxGraphArgs,
    operator: str,
    operand_id: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        operator=operator,
        operand_id=operand_id,
        label_type="UnaryExpression",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(operand_id)),
        label_ast="AST",
    )

    return args.n_id



from blends.models import (
    NId,
)
from blends.syntax.builders.string_literal import (
    build_string_literal_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    value = node_to_str(args.ast_graph, args.n_id)
    return build_string_literal_node(args, value)



from collections.abc import (
    Callable,
    Iterator,
)
from typing import (
    NamedTuple,
)

from blends.models import (
    NId,
    SyntaxGraph,
)

# Bool value indicates whether the founded node is a definition or not
SearchResult = tuple[bool, NId]


class SearchArgs(NamedTuple):
    graph: SyntaxGraph
    n_id: NId
    symbol: str
    def_only: bool


Searcher = Callable[[SearchArgs], Iterator[SearchResult]]



import tempfile
from itertools import count
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest
from tree_sitter import Node, Point

from blends.ast.build_graph import (
    _build_ast_graph,
    _has_content_node,
    _hash_node,
    _is_node_terminal,
    get_ast_graph_from_path,
)
from blends.models import Content, Graph, Language

if TYPE_CHECKING:
    from collections.abc import Iterator


EXPECTED_NODES_WITH_CHILDREN = 3
EXPECTED_EDGES_WITH_CHILDREN = 2
EXPECTED_LABEL_L_OFFSET = "6"
EXPECTED_LABEL_C_OFFSET = "11"


def create_mock_node(  # noqa: PLR0913
    node_type: str,
    *,
    start_point: tuple[int, int] = (0, 0),
    end_point: tuple[int, int] = (1, 0),
    start_byte: int = 0,
    end_byte: int = 10,
    children: list[Node] | None = None,
    has_error: bool = False,
) -> Node:
    node = Mock(spec=Node)
    node.type = node_type
    node.start_point = Point(*start_point)
    node.end_point = Point(*end_point)
    node.start_byte = start_byte
    node.end_byte = end_byte
    node.children = children or []
    node.has_error = has_error
    node.child_by_field_name = Mock(return_value=None)
    return node


@pytest.mark.blends_test_group("ast_unittesting")
def test_hash_node_consistency() -> None:
    node1 = create_mock_node("function_definition", start_point=(0, 0), end_point=(5, 10))
    node2 = create_mock_node("function_definition", start_point=(0, 0), end_point=(5, 10))

    hash1 = _hash_node(node1)
    hash2 = _hash_node(node2)

    assert hash1 == hash2

    node1 = create_mock_node("function_definition", start_point=(0, 0), end_point=(5, 10))
    node2 = create_mock_node("class_definition", start_point=(0, 0), end_point=(5, 10))

    hash1 = _hash_node(node1)
    hash2 = _hash_node(node2)

    assert hash1 != hash2

    node1 = create_mock_node("function_definition", start_point=(0, 0), end_point=(5, 10))
    node2 = create_mock_node("function_definition", start_point=(1, 0), end_point=(5, 10))

    hash1 = _hash_node(node1)
    hash2 = _hash_node(node2)

    assert hash1 != hash2

    node1 = create_mock_node("function_definition", start_point=(0, 0), end_point=(5, 10))
    node2 = create_mock_node("function_definition", start_point=(0, 0), end_point=(6, 10))

    hash1 = _hash_node(node1)
    hash2 = _hash_node(node2)

    assert hash1 != hash2


@pytest.mark.blends_test_group("ast_unittesting")
def test_is_node_terminal_consistency() -> None:
    node = create_mock_node("flow_mapping")
    assert _is_node_terminal(node, Language.YAML) is True

    node = create_mock_node("single_quote_scalar")
    assert _is_node_terminal(node, Language.YAML) is True

    node = create_mock_node("double_quote_scalar")
    assert _is_node_terminal(node, Language.YAML) is True

    node = create_mock_node("bang")
    assert _is_node_terminal(node, Language.SWIFT) is True

    node = create_mock_node("function_definition")
    assert _is_node_terminal(node, Language.PYTHON) is False
    assert _is_node_terminal(node, Language.JAVASCRIPT) is False
    assert _is_node_terminal(node, Language.JAVA) is False
    assert _is_node_terminal(node, Language.CSHARP) is False
    assert _is_node_terminal(node, Language.GO) is False

    node = create_mock_node("mapping")
    assert _is_node_terminal(node, Language.YAML) is False


@pytest.mark.blends_test_group("ast_unittesting")
def test_has_content_node_consistency() -> None:
    true_cases = [
        (Language.CSHARP, "character_literal"),
        (Language.CSHARP, "predefined_type"),
        (Language.CSHARP, "string_literal"),
        (Language.CSHARP, "verbatim_string_literal"),
        (Language.DART, "list_literal"),
        (Language.DART, "set_or_map_literal"),
        (Language.DART, "string_literal"),
        (Language.GO, "interpreted_string_literal"),
        (Language.JAVASCRIPT, "template_string"),
        (Language.JAVASCRIPT, "regex"),
        (Language.TYPESCRIPT, "template_string"),
        (Language.TYPESCRIPT, "regex"),
        (Language.KOTLIN, "string_literal"),
        (Language.SWIFT, "bang"),
        (Language.YAML, "single_quote_scalar"),
        (Language.YAML, "double_quote_scalar"),
        (Language.YAML, "flow_mapping"),
    ]

    false_cases = [
        (Language.CSHARP, "function_definition"),
        (Language.CSHARP, "class_declaration"),
        (Language.CSHARP, "template_string"),
        (Language.DART, "function_declaration"),
        (Language.DART, "class_definition"),
        (Language.DART, "character_literal"),
        (Language.GO, "function_declaration"),
        (Language.GO, "string_literal"),
        (Language.GO, "template_string"),
        (Language.JAVASCRIPT, "function_declaration"),
        (Language.JAVASCRIPT, "string_literal"),
        (Language.JAVASCRIPT, "class_declaration"),
        (Language.TYPESCRIPT, "function_declaration"),
        (Language.TYPESCRIPT, "string_literal"),
        (Language.TYPESCRIPT, "class_declaration"),
        (Language.KOTLIN, "function_declaration"),
        (Language.KOTLIN, "class_declaration"),
        (Language.KOTLIN, "template_string"),
        (Language.SWIFT, "function_declaration"),
        (Language.SWIFT, "string_literal"),
        (Language.SWIFT, "class_declaration"),
        (Language.YAML, "mapping"),
        (Language.YAML, "sequence"),
        (Language.YAML, "function_definition"),
        (Language.PYTHON, "string_literal"),
        (Language.PYTHON, "function_definition"),
        (Language.PYTHON, "template_string"),
        (Language.JAVA, "string_literal"),
        (Language.JAVA, "function_definition"),
        (Language.JAVA, "class_declaration"),
        (Language.PHP, "string_literal"),
        (Language.PHP, "function_definition"),
        (Language.RUBY, "string_literal"),
        (Language.RUBY, "function_definition"),
        (Language.SCALA, "string_literal"),
        (Language.SCALA, "function_definition"),
        (Language.HCL, "string_literal"),
        (Language.HCL, "function_definition"),
        (Language.JSON, "string_literal"),
        (Language.JSON, "function_definition"),
    ]

    for language, node_type in true_cases:
        node = create_mock_node(node_type)
        assert _has_content_node(node, language) is True, (
            f"Expected True for {language.value} with node type {node_type}"
        )

    for language, node_type in false_cases:
        node = create_mock_node(node_type)
        assert _has_content_node(node, language) is False, (
            f"Expected False for {language.value} with node type {node_type}"
        )


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_single_node() -> None:
    content = Content(b"test", "test", Language.PYTHON, Path("test.py"))
    node = create_mock_node("module", start_byte=0, end_byte=4)
    counter = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, node, counter, graph)

    assert len(result.nodes) == 1
    assert "1" in result.nodes
    assert result.nodes["1"]["label_type"] == "module"
    assert result.nodes["1"]["label_l"] == "1"
    assert result.nodes["1"]["label_c"] == "1"


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_with_children() -> None:
    child1 = create_mock_node(
        "child1", start_byte=0, end_byte=2, start_point=(0, 0), end_point=(0, 2)
    )
    child2 = create_mock_node(
        "child2", start_byte=3, end_byte=5, start_point=(0, 3), end_point=(0, 5)
    )
    parent = create_mock_node("parent", children=[child1, child2], start_byte=0, end_byte=5)

    content = Content(b"test", "test", Language.PYTHON, Path("test.py"))
    counter = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, parent, counter, graph)

    assert len(result.nodes) == EXPECTED_NODES_WITH_CHILDREN
    assert len(result.edges) == EXPECTED_EDGES_WITH_CHILDREN


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_terminal_node_has_text() -> None:
    content = Content(b"hello", "hello", Language.PYTHON, Path("test.py"))
    node = create_mock_node("string", start_byte=0, end_byte=5, children=[])
    counter: Iterator[str] = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, node, counter, graph)

    assert "label_text" in result.nodes["1"]
    assert result.nodes["1"]["label_text"] == "hello"


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_non_terminal_node_no_text() -> None:
    child = create_mock_node("child", start_byte=0, end_byte=2, children=[])
    parent = create_mock_node("parent", children=[child], start_byte=0, end_byte=2)

    content = Content(b"ab", "ab", Language.PYTHON, Path("test.py"))
    counter = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, parent, counter, graph)

    assert "label_text" not in result.nodes["1"]


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_edges_have_correct_labels() -> None:
    child = create_mock_node("child", start_byte=0, end_byte=2, children=[])
    parent = create_mock_node("parent", children=[child], start_byte=0, end_byte=2)

    content = Content(b"ab", "ab", Language.PYTHON, Path("test.py"))
    counter = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, parent, counter, graph)

    edges = list(result.edges(data=True))
    assert len(edges) == 1
    edge_data = edges[0][2]
    assert edge_data["label_ast"] == "AST"
    assert edge_data["label_index"] == "0"


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_node_attributes() -> None:
    content = Content(b"test", "test", Language.PYTHON, Path("test.py"))
    node = create_mock_node(
        "function_definition", start_point=(5, 10), end_point=(10, 20), start_byte=0, end_byte=4
    )
    counter: Iterator[str] = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, node, counter, graph)

    assert result.nodes["1"]["label_l"] == EXPECTED_LABEL_L_OFFSET
    assert result.nodes["1"]["label_c"] == EXPECTED_LABEL_C_OFFSET
    assert result.nodes["1"]["label_type"] == "function_definition"


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_not_implemented_error() -> None:
    content = Content(b"test", "test", Language.PYTHON, Path("test.py"))
    invalid_node = "not a node"
    counter = map(str, count(1))
    graph = Graph()

    with pytest.raises(NotImplementedError):
        _build_ast_graph(content, invalid_node, counter, graph)  # type: ignore[arg-type]


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_yaml_terminal_node() -> None:
    content = Content(b"value", "value", Language.YAML, Path("test.yaml"))
    node = create_mock_node("single_quote_scalar", start_byte=0, end_byte=5, children=[])
    counter: Iterator[str] = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, node, counter, graph)

    assert "label_text" in result.nodes["1"]
    assert result.nodes["1"]["label_text"] == "value"


@pytest.mark.blends_test_group("ast_unittesting")
def test_build_ast_graph_has_content_node_includes_text() -> None:
    content = Content(b"hello", "hello", Language.JAVASCRIPT, Path("test.js"))
    child = create_mock_node("child_node", start_byte=0, end_byte=2, children=[])
    node = create_mock_node("template_string", start_byte=0, end_byte=5, children=[child])
    counter = map(str, count(1))
    graph = Graph()

    result = _build_ast_graph(content, node, counter, graph)

    assert "label_text" in result.nodes["1"]
    assert result.nodes["1"]["label_text"] == "hello"


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_python_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.py"
        test_content = "def hello():\n    return 'world'"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_javascript_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.js"
        test_content = "function test() { return true; }"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_go_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.go"
        test_content = "package main\n\nfunc main() {}"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_yaml_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.yaml"
        test_content = "key: value\nother: data"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_nonexistent_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        non_existent_file = Path(temp_dir) / "nonexistent.py"

        result = get_ast_graph_from_path(non_existent_file)

        assert result is None


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_unsupported_language() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.unknown"
        test_file.write_text("some content", encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is None


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_empty_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "empty.py"
        test_file.touch()

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_java_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "Test.java"
        test_content = "public class Test { }"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_typescript_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.ts"
        test_content = "const x: number = 42;"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0


@pytest.mark.blends_test_group("ast_unittesting")
def test_get_ast_graph_from_path_valid_csharp_file() -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "Test.cs"
        test_content = "public class Test { }"
        test_file.write_text(test_content, encoding="utf-8")

        result = get_ast_graph_from_path(test_file)

        assert result is not None
        assert isinstance(result, Graph)
        assert len(result.nodes) > 0



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_await_expression_node(
    args: SyntaxGraphArgs,
    expression: NId,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        expression_id=expression,
        label_type="AwaitExpression",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(expression)),
        label_ast="AST",
    )

    return args.n_id



import pytest

from test.stack.test_helpers import (
    count_defs_refs,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("stack_unittesting")
def test_same_function_name_different_scopes() -> None:
    view, record = setup_stack_resolution("case_same_name_diff_scope.py")
    defs, _refs = count_defs_refs(view, "helper")
    assert defs >= 2, f"Expected at least 2 defs of 'helper', got {defs}"  # noqa: PLR2004

    result_0 = resolve_symbol(view, record, "helper", ref_index=0)
    result_1 = resolve_symbol(view, record, "helper", ref_index=1)
    assert len(result_0) == 1, f"Ref in outer() should resolve to 1 def, got {len(result_0)}"
    assert len(result_1) == 1, f"Ref in another() should resolve to 1 def, got {len(result_1)}"
    assert result_0 != result_1, "Should resolve to different definitions"


@pytest.mark.blends_test_group("stack_unittesting")
def test_sibling_function_parameter_isolation() -> None:
    view, record = setup_stack_resolution("case_sibling_param_scope.py")
    defs, _refs = count_defs_refs(view, "x")
    assert defs >= 2  # noqa: PLR2004

    result_0 = resolve_symbol(view, record, "x", ref_index=0)
    assert len(result_0) == 1, f"'x' in compute() should resolve to 1 def, got {len(result_0)}"


@pytest.mark.blends_test_group("stack_unittesting")
def test_module_level_visible_from_function() -> None:
    view, record = setup_stack_resolution("case_func_arg.py")
    result = resolve_symbol(view, record, "value")
    assert len(result) >= 1, "Module-level 'value' should be visible from function call"



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group,
)
from blends.syntax.builders.method_invocation import (
    DirectChildren,
    build_method_invocation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    expr_id = adj_ast(graph, args.n_id)[0]
    expr = node_to_str(graph, expr_id)

    matches = match_ast_group(graph, args.n_id, "value_arguments", "annotated_lambda")

    children: DirectChildren = {"expression_id": expr_id}
    if value_arguments := matches.get("value_arguments"):
        children["arguments_id"] = value_arguments[0]

    if (annotated_lambda := matches.get("annotated_lambda")) and (
        lambda_children := adj_ast(graph, annotated_lambda[0], label_type="lambda_literal")
    ):
        children["object_id"] = lambda_children[0]

    return build_method_invocation_node(args, expr, children)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_group_d,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = match_ast_group_d(args.ast_graph, args.n_id, "argument") or match_ast_group_d(
        args.ast_graph,
        args.n_id,
        "attribute_argument",
    )

    args_ids = []
    for _id in c_ids:
        if (childs := adj_ast(graph, _id)) and len(childs) == 1:
            args_ids.append(childs[0])
        else:
            args_ids.append(_id)

    return build_argument_list_node(args, iter(args_ids))



import pytest

from blends.stack.node_kinds import StackGraphNodeKind
from test.stack.test_helpers import (
    count_defs_refs,
    find_symbol_nodes,
    resolve_symbol,
    setup_stack_resolution,
)


@pytest.mark.blends_test_group("functional")
def test_self_attribute_resolves_to_class_attribute() -> None:
    view, record = setup_stack_resolution("case_self_attr.py")
    defs, refs = count_defs_refs(view, "value")
    result = resolve_symbol(view, record, "value")

    assert defs >= 1, f"Expected at least 1 def of 'value', got {defs}"
    assert refs >= 1, f"Expected at least 1 ref to 'value', got {refs}"
    assert len(result) >= 1, f"Expected self.value to resolve to class attr, got {len(result)}"


@pytest.mark.blends_test_group("functional")
def test_self_attr_does_not_resolve_to_inner_class_attribute() -> None:
    view, record = setup_stack_resolution("case_self_attr_nested_class.py")
    defs, refs = count_defs_refs(view, "value")
    result = resolve_symbol(view, record, "value")

    assert defs >= 2, (  # noqa: PLR2004
        f"Expected at least 2 defs of 'value' (outer + inner), got {defs}"
    )
    assert refs >= 1, f"Expected at least 1 ref to 'value', got {refs}"
    assert len(result) == 1, (
        f"self.value in outer method should resolve only to outer class attr, got {len(result)}"
    )

    symbol_id = view.symbols.symbol_to_id.get("value")
    assert symbol_id is not None
    def_nodes = find_symbol_nodes(view, symbol_id=symbol_id, is_definition=True)
    outer_value_def = next(
        idx for idx in def_nodes if view.node_kind[idx] == StackGraphNodeKind.POP_SYMBOL.value
    )
    assert result == [outer_value_def], (
        f"Expected self.value to resolve to Outer.value definition, "
        f"got {result}, expected [{outer_value_def}]"
    )


@pytest.mark.blends_test_group("functional")
def test_self_attr_resolves_with_typed_self_parameter() -> None:
    view, record = setup_stack_resolution("case_self_attr_typed_self.py")
    defs, refs = count_defs_refs(view, "value")
    result = resolve_symbol(view, record, "value")

    assert defs >= 1, f"Expected at least 1 def of 'value', got {defs}"
    assert refs >= 1, f"Expected at least 1 ref to 'value', got {refs}"
    assert len(result) >= 1, (
        f"Typed self parameter should not prevent self.value resolution, got {len(result)}"
    )


@pytest.mark.blends_test_group("functional")
def test_typed_param_member_access_resolves_to_class_attribute() -> None:
    view, record = setup_stack_resolution("case_typed_param_member_access.py")
    defs, refs = count_defs_refs(view, "value")

    assert defs >= 1, f"Expected at least 1 def of 'value', got {defs}"
    assert refs >= 2, f"Expected at least 2 refs to 'value', got {refs}"  # noqa: PLR2004

    results = [resolve_symbol(view, record, "value", ref_index=i) for i in range(refs)]

    assert all(len(r) >= 1 for r in results), (
        f"Every value ref should resolve to the class attribute, got {[len(r) for r in results]}"
    )
    unique_resolutions = {tuple(sorted(r)) for r in results}
    assert len(unique_resolutions) == 1, (
        f"All value refs must resolve to the same definition, got {unique_resolutions}"
    )



from unittest.mock import (
    patch,
)

import pytest

from blends.models import (
    Graph,
    Language,
    NId,
    SyntaxGraph,
)
from blends.stack.attributes import STACK_GRAPH_KIND, STACK_GRAPH_SYMBOL
from blends.stack.node_kinds import (
    StackGraphNodeKind,
)
from blends.syntax.builders.member_access import (
    build_member_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
    SyntaxMetadata,
)


@pytest.fixture  # type: ignore[misc,attr-defined]
def mock_args() -> SyntaxGraphArgs:  # type: ignore[misc]
    syntax_graph = SyntaxGraph()
    ast_graph = Graph()
    metadata: SyntaxMetadata = {
        "class_path": [],
        "scope_stack": ["1"],
        "next_synthetic_id": 100,
    }
    syntax_graph.add_node("1", label_type="File")
    syntax_graph.add_node("5", label_type="SymbolLookup", symbol="obj")

    def generic(args: SyntaxGraphArgs) -> NId:
        return args.n_id

    return SyntaxGraphArgs(
        generic=generic,
        path="test.py",
        language=Language.PYTHON,
        ast_graph=ast_graph,
        syntax_graph=syntax_graph,
        n_id="2",
        metadata=metadata,
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_member_access_creates_push_symbol_for_member(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_member_access_node(
            mock_args,
            member="attr",
            expression="obj.attr",
            expression_id="5",
        )

    assert (
        mock_args.syntax_graph.nodes["2"][STACK_GRAPH_KIND] == StackGraphNodeKind.PUSH_SYMBOL.value
    )
    assert mock_args.syntax_graph.nodes["2"][STACK_GRAPH_SYMBOL] == "attr"


@pytest.mark.blends_test_group("stack_unittesting")
def test_member_access_creates_dot_separator_node(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        build_member_access_node(
            mock_args,
            member="attr",
            expression="obj.attr",
            expression_id="5",
        )

    graph = mock_args.syntax_graph
    dot_nodes = [
        s
        for s in graph.successors("2")
        if graph.nodes[s].get(STACK_GRAPH_KIND) == StackGraphNodeKind.PUSH_SYMBOL.value
        and graph.nodes[s].get(STACK_GRAPH_SYMBOL) == "."
    ]
    assert len(dot_nodes) == 1
    dot_node = dot_nodes[0]
    assert graph.has_edge(dot_node, "5")


@pytest.mark.blends_test_group("stack_unittesting")
def test_member_access_no_stack_graph_when_feature_disabled(
    mock_args: SyntaxGraphArgs,
) -> None:
    with patch("blends.ctx.ctx.has_feature_flag", return_value=False):
        build_member_access_node(
            mock_args,
            member="attr",
            expression="obj.attr",
            expression_id="5",
        )

    assert STACK_GRAPH_KIND not in mock_args.syntax_graph.nodes["2"]
    assert STACK_GRAPH_SYMBOL not in mock_args.syntax_graph.nodes["2"]



import logging
from collections.abc import Callable

from blends.models import (
    GraphDB,
    NId,
    SyntaxGraph,
)
from blends.symbolic_evaluation.cases import (
    argument,
    argument_list,
    assignment,
    binary_operation,
    element_access,
    else_clause,
    execution_block,
    for_each_statement,
    for_statement,
    general_evaluator,
    if_statement,
    import_statement,
    literal,
    member_access,
    method_declaration,
    method_invocation,
    named_argument,
    new_expression,
    not_dangerous,
    object_creation,
    object_node,
    pair,
    parameter,
    parenthesized_expression,
    return_node,
    spread_element,
    symbol_lookup,
    ternary_operation,
    try_statement,
    unary_expression,
    using_statement,
    variable_declaration,
    while_statement,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
    register_evaluators,
)
from blends.symbolic_evaluation.models import (
    MissingSymbolicEvalError,
    Path,
    SymbolicEvalArgs,
    SymbolicEvaluation,
)
from blends.symbolic_evaluation.utils import get_backward_paths

LOGGER = logging.getLogger(__name__)

register_evaluators(
    {
        "Annotation": not_dangerous.evaluate,
        "Argument": argument.evaluate,
        "ArgumentList": argument_list.evaluate,
        "ArrayInitializer": general_evaluator.evaluate,
        "Assignment": assignment.evaluate,
        "Attribute": not_dangerous.evaluate,
        "AwaitExpression": general_evaluator.evaluate,
        "BinaryOperation": binary_operation.evaluate,
        "Break": not_dangerous.evaluate,
        "CatchClause": not_dangerous.evaluate,
        "CatchDeclaration": not_dangerous.evaluate,
        "Comment": not_dangerous.evaluate,
        "Continue": not_dangerous.evaluate,
        "Class": not_dangerous.evaluate,
        "ClassBody": not_dangerous.evaluate,
        "Debugger": not_dangerous.evaluate,
        "DeclarationBlock": not_dangerous.evaluate,
        "DoStatement": not_dangerous.evaluate,
        "ElementAccess": element_access.evaluate,
        "ElseClause": else_clause.evaluate,
        "ExecutionBlock": execution_block.evaluate,
        "Export": not_dangerous.evaluate,
        "ExpressionStatement": general_evaluator.evaluate,
        "File": not_dangerous.evaluate,
        "FinallyClause": not_dangerous.evaluate,
        "ForEachStatement": for_each_statement.evaluate,
        "ForStatement": for_statement.evaluate,
        "If": if_statement.evaluate,
        "Import": import_statement.evaluate,
        "JsxElement": not_dangerous.evaluate,
        "Literal": literal.evaluate,
        "MemberAccess": member_access.evaluate,
        "MethodDeclaration": method_declaration.evaluate,
        "MethodInvocation": method_invocation.evaluate,
        "MissingNode": not_dangerous.evaluate,
        "Modifiers": general_evaluator.evaluate,
        "ModuleImport": not_dangerous.evaluate,
        "NamedArgument": named_argument.evaluate,
        "Namespace": not_dangerous.evaluate,
        "NewExpression": new_expression.evaluate,
        "Object": object_node.evaluate,
        "ObjectCreation": object_creation.evaluate,
        "Pair": pair.evaluate,
        "Parameter": parameter.evaluate,
        "ParameterList": general_evaluator.evaluate,
        "ParenthesizedExpression": parenthesized_expression.evaluate,
        "ReservedWord": not_dangerous.evaluate,
        "RestPattern": not_dangerous.evaluate,
        "Return": return_node.evaluate,
        "Selector": not_dangerous.evaluate,
        "SpreadElement": spread_element.evaluate,
        "SwitchBody": general_evaluator.evaluate,
        "SwitchSection": general_evaluator.evaluate,
        "SwitchStatement": general_evaluator.evaluate,
        "SymbolLookup": symbol_lookup.evaluate,
        "TernaryOperation": ternary_operation.evaluate,
        "TryStatement": try_statement.evaluate,
        "This": not_dangerous.evaluate,
        "ThrowStatement": general_evaluator.evaluate,
        "TypeOf": not_dangerous.evaluate,
        "UnaryExpression": unary_expression.evaluate,
        "UsingStatement": using_statement.evaluate,
        "VariableDeclaration": variable_declaration.evaluate,
        "WhileStatement": while_statement.evaluate,
        "Yield": not_dangerous.evaluate,
    }
)


def evaluate(
    graph: SyntaxGraph,
    path: Path,
    n_id: NId,
    graph_db: GraphDB | None = None,
    method_evaluators: dict[str, Callable[[SymbolicEvalArgs], SymbolicEvaluation]] | None = None,
) -> SymbolicEvaluation | None:
    try:
        evaluation: dict[NId, bool] = {}
        return generic(
            SymbolicEvalArgs(
                evaluation,
                graph,
                path,
                n_id,
                set(),
                graph_db,
                method_evaluators,
            ),
        )
    except MissingSymbolicEvalError:
        LOGGER.exception("Missing symbolic evaluator in %s", path)
        return None


def get_node_evaluation_results(  # noqa: PLR0913
    graph: SyntaxGraph,
    n_id: NId,
    triggers_goal: set[str],
    *,
    danger_goal: bool = True,
    graph_db: GraphDB | None = None,
    method_evaluators: dict[str, Callable[[SymbolicEvalArgs], SymbolicEvaluation]] | None = None,
) -> bool:
    for path in get_backward_paths(graph, n_id):
        evaluation = evaluate(graph, path, n_id, graph_db, method_evaluators)
        if evaluation and evaluation.danger == danger_goal and evaluation.triggers == triggers_goal:
            return True
    return False


def get_node_evaluation_results_advanced(
    graph: SyntaxGraph,
    n_id: NId,
    evaluation_goal: Callable[[bool | None, set[str]], bool],
    method_evaluators: dict[str, Callable[[SymbolicEvalArgs], SymbolicEvaluation]] | None = None,
    graph_db: GraphDB | None = None,
) -> bool:
    for path in get_backward_paths(graph, n_id):
        evaluation = evaluate(graph, path, n_id, graph_db, method_evaluators=method_evaluators)
        if evaluation and evaluation_goal(evaluation.danger, evaluation.triggers):
            return True
    return False



import logging
from pathlib import Path

from blends.content.custom_parsers import custom_parsers_by_language
from blends.models import EXTENSION_TO_LANGUAGE, Content, Language

LOGGER = logging.getLogger(__name__)

MAX_FILE_SIZE: int = 1024 * 500


class FileTooLargeError(Exception):
    pass


def get_language_by_path(path: Path) -> Language:
    return EXTENSION_TO_LANGUAGE.get(path.suffix.lower(), Language.NOT_SUPPORTED)


def _get_file_raw_content(file_path: Path, size: int = MAX_FILE_SIZE) -> bytes:
    file_size = file_path.stat().st_size
    if file_size > size:
        raise FileTooLargeError(str(file_path))

    with file_path.open("rb") as handle:
        return handle.read(size)


def get_content_by_path(
    path: Path,
    max_size: int = MAX_FILE_SIZE,
) -> Content | None:
    """Interface function for the content module.

    Loads and returns file content for a given path.
    Accepts a Path to a file and an optional maximum file size in bytes.
    Reads the file up to the specified size, decodes to string, and wraps both bytes and string
    in a Content object. Applies language-aware custom parsers and validators if applicable,
    returning parsed or validated content. Returns None if the file cannot be read or is too large.

    Expects:
        path: Path to the file to be read.
        max_size: Maximum allowed file size in bytes (default: 512 KB).

    Returns:
        Content object with bytes, string, and file size, or None if reading fails or file is too
        large.

    """
    content = None
    language = get_language_by_path(path)
    try:
        content_bytes = _get_file_raw_content(path, max_size)
        content_str = content_bytes.decode("utf-8", errors="ignore")
        content = Content(content_bytes, content_str, language, path)

    except FileTooLargeError:
        LOGGER.warning("File too large: %s, ignoring", path)
        return None
    except (FileNotFoundError, PermissionError, OSError, IsADirectoryError):
        content = None
        LOGGER.exception("Error reading file: %s", path)
        return None
    if content and (language_parsers := custom_parsers_by_language.get(language)):
        for custom_parser in language_parsers:
            if custom_parser.validator(content) and (
                fixed_content := custom_parser.parser(content)
            ):
                return fixed_content

    return content



from collections.abc import (
    Iterable,
)

from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_parameter_list_node(args: SyntaxGraphArgs, c_ids: Iterable[NId]) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="ParameterList",
    )

    for c_id in c_ids:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(c_id)),
            label_ast="AST",
        )

    return args.n_id



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path.partial_path import PartialPath
from blends.stack.partial_path.partial_stacks import (
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
)
from blends.stack.partial_path.variables import (
    ScopeStackVariable,
    SymbolStackVariable,
)

if TYPE_CHECKING:
    from blends.stack.view import StackGraphView


@dataclass(slots=True)
class StackConditions:
    symbol_pre: PartialSymbolStack
    symbol_post: PartialSymbolStack
    scope_pre: PartialScopeStack
    scope_post: PartialScopeStack


def default_stacks() -> tuple[
    PartialSymbolStack, PartialSymbolStack, PartialScopeStack, PartialScopeStack
]:
    symbol_var = PartialSymbolStack.from_variable(SymbolStackVariable.initial())
    scope_var = PartialScopeStack.from_variable(ScopeStackVariable.initial())
    return symbol_var, symbol_var, scope_var, scope_var


def apply_node_semantics(view: StackGraphView, node_index: int) -> StackConditions:
    symbol_pre, symbol_post, scope_pre, scope_post = default_stacks()
    stacks = StackConditions(
        symbol_pre=symbol_pre,
        symbol_post=symbol_post,
        scope_pre=scope_pre,
        scope_post=scope_post,
    )
    handler = _HANDLERS.get(view.kind_at(node_index))
    if handler is not None:
        stacks = handler(view, node_index, stacks)
    return stacks


def create_seed_path(view: StackGraphView, seed_node_index: int) -> PartialPath:
    stacks = apply_node_semantics(view, seed_node_index)
    return PartialPath(
        start_node_index=seed_node_index,
        end_node_index=seed_node_index,
        symbol_stack_precondition=stacks.symbol_pre,
        symbol_stack_postcondition=stacks.symbol_post,
        scope_stack_precondition=stacks.scope_pre,
        scope_stack_postcondition=stacks.scope_post,
        edges=(),
    )


def _handle_push_symbol(
    view: StackGraphView,
    sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    symbol_id = view.symbol_id_at(sink_node_index)
    if symbol_id is None:
        return stacks
    stacks.symbol_post = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=None),),
        variable=SymbolStackVariable.initial(),
    )
    return stacks


def _handle_pop_symbol(
    view: StackGraphView,
    sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    symbol_id = view.symbol_id_at(sink_node_index)
    if symbol_id is None:
        return stacks
    stacks.symbol_pre = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=None),),
        variable=SymbolStackVariable.initial(),
    )
    stacks.symbol_post = PartialSymbolStack.from_variable(SymbolStackVariable.initial())
    return stacks


def _handle_push_scoped_symbol(
    view: StackGraphView,
    sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    symbol_id = view.symbol_id_at(sink_node_index)
    scope_index = view.scope_index_at(sink_node_index)
    if symbol_id is None or scope_index is None:
        return stacks
    attached_scopes = PartialScopeStack(
        scopes=(scope_index,), variable=ScopeStackVariable.initial()
    )
    stacks.symbol_post = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=attached_scopes),),
        variable=SymbolStackVariable.initial(),
    )
    return stacks


def _handle_pop_scoped_symbol(
    view: StackGraphView,
    sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    symbol_id = view.symbol_id_at(sink_node_index)
    if symbol_id is None:
        return stacks
    attached_scopes = PartialScopeStack(scopes=(), variable=ScopeStackVariable.initial())
    stacks.symbol_pre = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=symbol_id, scopes=attached_scopes),),
        variable=SymbolStackVariable.initial(),
    )
    stacks.symbol_post = PartialSymbolStack.from_variable(SymbolStackVariable.initial())
    stacks.scope_post = attached_scopes
    return stacks


def _handle_drop_scopes(
    _view: StackGraphView,
    _sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    stacks.scope_pre = PartialScopeStack.from_variable(ScopeStackVariable.initial())
    stacks.scope_post = PartialScopeStack.empty()
    return stacks


def _handle_jump_to(
    _view: StackGraphView,
    _sink_node_index: int,
    stacks: StackConditions,
) -> StackConditions:
    stacks.scope_pre = PartialScopeStack.from_variable(ScopeStackVariable.initial())
    stacks.scope_post = PartialScopeStack.from_variable(ScopeStackVariable.initial())
    return stacks


_HANDLERS = {
    StackGraphNodeKind.PUSH_SYMBOL.value: _handle_push_symbol,
    StackGraphNodeKind.POP_SYMBOL.value: _handle_pop_symbol,
    StackGraphNodeKind.PUSH_SCOPED_SYMBOL.value: _handle_push_scoped_symbol,
    StackGraphNodeKind.POP_SCOPED_SYMBOL.value: _handle_pop_scoped_symbol,
    StackGraphNodeKind.DROP_SCOPES.value: _handle_drop_scopes,
    StackGraphNodeKind.JUMP_TO.value: _handle_jump_to,
}



from dataclasses import dataclass
from typing import TYPE_CHECKING

from blends.stack.partial_path.bindings import _SymbolUnifyBindings
from blends.stack.partial_path.errors import (
    _raise_scope_stack_unsatisfied,
    _raise_symbol_stack_unsatisfied,
)

if TYPE_CHECKING:
    from blends.stack.partial_path.bindings import (
        PartialScopeStackBindings,
        PartialSymbolStackBindings,
    )
    from blends.stack.partial_path.variables import (
        ScopeStackVariable,
        SymbolStackVariable,
    )


def _common_prefix_len_scopes(lhs: tuple[int, ...], rhs: tuple[int, ...]) -> int:
    prefix_len = 0
    for lhs_scope, rhs_scope in zip(lhs, rhs, strict=False):
        if lhs_scope != rhs_scope:
            _raise_scope_stack_unsatisfied()
        prefix_len += 1
    return prefix_len


def _unify_scope_suffix_empty(
    lhs: "PartialScopeStack",
    rhs: "PartialScopeStack",
    bindings: "PartialScopeStackBindings",
) -> "PartialScopeStack":
    lhs_var = lhs.variable
    rhs_var = rhs.variable
    if lhs_var is None and rhs_var is None:
        return lhs
    if lhs_var is None and rhs_var is not None:
        bindings.add(rhs_var, PartialScopeStack.empty())
        return rhs
    if lhs_var is not None and rhs_var is None:
        bindings.add(lhs_var, PartialScopeStack.empty())
        return lhs
    if lhs_var is None or rhs_var is None:
        _raise_scope_stack_unsatisfied()
    if lhs_var == rhs_var:
        _raise_scope_stack_unsatisfied()
    bindings.add(rhs_var, PartialScopeStack.from_variable(lhs_var))
    return lhs


def _unify_scope_rhs_is_prefix_of_lhs(
    lhs: "PartialScopeStack",
    rhs: "PartialScopeStack",
    lhs_suffix_scopes: tuple[int, ...],
    bindings: "PartialScopeStackBindings",
) -> "PartialScopeStack":
    rhs_var = rhs.variable
    if rhs_var is None:
        _raise_scope_stack_unsatisfied()
    if lhs.variable is not None and lhs.variable == rhs_var:
        _raise_scope_stack_unsatisfied()
    bindings.add(rhs_var, PartialScopeStack(scopes=lhs_suffix_scopes, variable=lhs.variable))
    return lhs


def _unify_scope_lhs_is_prefix_of_rhs(
    lhs: "PartialScopeStack",
    rhs: "PartialScopeStack",
    rhs_suffix_scopes: tuple[int, ...],
    bindings: "PartialScopeStackBindings",
) -> "PartialScopeStack":
    lhs_var = lhs.variable
    if lhs_var is None:
        _raise_scope_stack_unsatisfied()
    if rhs.variable is not None and rhs.variable == lhs_var:
        _raise_scope_stack_unsatisfied()
    bindings.add(lhs_var, PartialScopeStack(scopes=rhs_suffix_scopes, variable=rhs.variable))
    return rhs


def _unify_symbol_suffix_empty(
    prefix: tuple["PartialScopedSymbol", ...],
    lhs: "PartialSymbolStack",
    rhs: "PartialSymbolStack",
    symbol_bindings: "PartialSymbolStackBindings",
    scope_bindings: "PartialScopeStackBindings",
) -> "PartialSymbolStack":
    lhs_var = lhs.variable
    rhs_var = rhs.variable
    if lhs_var is None and rhs_var is None:
        return PartialSymbolStack(symbols=prefix, variable=None)
    if lhs_var is None and rhs_var is not None:
        symbol_bindings.add(rhs_var, PartialSymbolStack.empty(), scope_bindings)
        return PartialSymbolStack(symbols=prefix, variable=rhs_var)
    if lhs_var is not None and rhs_var is None:
        symbol_bindings.add(lhs_var, PartialSymbolStack.empty(), scope_bindings)
        return PartialSymbolStack(symbols=prefix, variable=lhs_var)
    if lhs_var is None or rhs_var is None:
        _raise_symbol_stack_unsatisfied()
    if lhs_var == rhs_var:
        _raise_symbol_stack_unsatisfied()
    symbol_bindings.add(rhs_var, PartialSymbolStack.from_variable(lhs_var), scope_bindings)
    return PartialSymbolStack(symbols=prefix, variable=lhs_var)


def _unify_symbol_rhs_is_prefix_of_lhs(
    prefix: tuple["PartialScopedSymbol", ...],
    lhs: "PartialSymbolStack",
    rhs: "PartialSymbolStack",
    lhs_suffix_symbols: tuple["PartialScopedSymbol", ...],
    bindings: _SymbolUnifyBindings,
) -> "PartialSymbolStack":
    rhs_var = rhs.variable
    if rhs_var is None:
        _raise_symbol_stack_unsatisfied()
    if lhs.variable is not None and lhs.variable == rhs_var:
        _raise_symbol_stack_unsatisfied()
    bindings.symbol_bindings.add(
        rhs_var,
        PartialSymbolStack(symbols=lhs_suffix_symbols, variable=lhs.variable),
        bindings.scope_bindings,
    )
    return PartialSymbolStack(symbols=prefix + lhs_suffix_symbols, variable=lhs.variable)


def _unify_symbol_lhs_is_prefix_of_rhs(
    prefix: tuple["PartialScopedSymbol", ...],
    lhs: "PartialSymbolStack",
    rhs: "PartialSymbolStack",
    rhs_suffix_symbols: tuple["PartialScopedSymbol", ...],
    bindings: _SymbolUnifyBindings,
) -> "PartialSymbolStack":
    lhs_var = lhs.variable
    if lhs_var is None:
        _raise_symbol_stack_unsatisfied()
    if rhs.variable is not None and rhs.variable == lhs_var:
        _raise_symbol_stack_unsatisfied()
    bindings.symbol_bindings.add(
        lhs_var,
        PartialSymbolStack(symbols=rhs_suffix_symbols, variable=rhs.variable),
        bindings.scope_bindings,
    )
    return PartialSymbolStack(symbols=prefix + rhs_suffix_symbols, variable=rhs.variable)


@dataclass(frozen=True, slots=True)
class PartialScopeStack:
    scopes: tuple[int, ...]
    variable: "ScopeStackVariable | None"

    @classmethod
    def empty(cls) -> "PartialScopeStack":
        return cls(scopes=(), variable=None)

    @classmethod
    def from_variable(cls, variable: "ScopeStackVariable") -> "PartialScopeStack":
        return cls(scopes=(), variable=variable)

    def has_variable(self) -> bool:
        return self.variable is not None

    def can_match_empty(self) -> bool:
        return len(self.scopes) == 0

    def can_only_match_empty(self) -> bool:
        return len(self.scopes) == 0 and self.variable is None

    def contains_scopes(self) -> bool:
        return len(self.scopes) > 0

    def with_offset(self, scope_offset: int) -> "PartialScopeStack":
        if self.variable is None or scope_offset == 0:
            return self
        return PartialScopeStack(
            scopes=self.scopes,
            variable=self.variable.with_offset(scope_offset),
        )

    def largest_scope_stack_var(self) -> int:
        if self.variable is None:
            return 0
        return self.variable.as_int()

    def apply_bindings(self, bindings: "PartialScopeStackBindings") -> "PartialScopeStack":
        if self.variable is None:
            return self
        bound = bindings.get(self.variable)
        if bound is None:
            return self
        return PartialScopeStack(
            scopes=self.scopes + bound.scopes,
            variable=bound.variable,
        )

    def unify(
        self, rhs: "PartialScopeStack", bindings: "PartialScopeStackBindings"
    ) -> "PartialScopeStack":
        prefix_len = _common_prefix_len_scopes(self.scopes, rhs.scopes)
        lhs_suffix_scopes = self.scopes[prefix_len:]
        rhs_suffix_scopes = rhs.scopes[prefix_len:]

        if len(lhs_suffix_scopes) == 0 and len(rhs_suffix_scopes) == 0:
            return _unify_scope_suffix_empty(self, rhs, bindings)
        if len(rhs_suffix_scopes) == 0:
            return _unify_scope_rhs_is_prefix_of_lhs(self, rhs, lhs_suffix_scopes, bindings)
        if len(lhs_suffix_scopes) == 0:
            return _unify_scope_lhs_is_prefix_of_rhs(self, rhs, rhs_suffix_scopes, bindings)
        _raise_scope_stack_unsatisfied()
        raise AssertionError  # unreachable


@dataclass(frozen=True, slots=True)
class PartialScopedSymbol:
    symbol_id: int
    scopes: "PartialScopeStack | None"

    def unify(
        self, rhs: "PartialScopedSymbol", bindings: "PartialScopeStackBindings"
    ) -> "PartialScopedSymbol":
        if self.symbol_id != rhs.symbol_id:
            _raise_symbol_stack_unsatisfied()
        if (self.scopes is None) != (rhs.scopes is None):
            _raise_symbol_stack_unsatisfied()
        if self.scopes is None or rhs.scopes is None:
            return self
        unified_scopes = self.scopes.unify(rhs.scopes, bindings)
        return PartialScopedSymbol(symbol_id=self.symbol_id, scopes=unified_scopes)

    def with_offset(self, scope_offset: int) -> "PartialScopedSymbol":
        if self.scopes is None:
            return self
        return PartialScopedSymbol(
            symbol_id=self.symbol_id,
            scopes=self.scopes.with_offset(scope_offset),
        )

    def largest_scope_stack_var(self) -> int:
        if self.scopes is None:
            return 0
        return self.scopes.largest_scope_stack_var()

    def apply_bindings(self, bindings: "PartialScopeStackBindings") -> "PartialScopedSymbol":
        if self.scopes is None:
            return self
        return PartialScopedSymbol(
            symbol_id=self.symbol_id,
            scopes=self.scopes.apply_bindings(bindings),
        )


@dataclass(frozen=True, slots=True)
class PartialSymbolStack:
    symbols: tuple["PartialScopedSymbol", ...]
    variable: "SymbolStackVariable | None"

    @classmethod
    def empty(cls) -> "PartialSymbolStack":
        return cls(symbols=(), variable=None)

    @classmethod
    def from_variable(cls, variable: "SymbolStackVariable") -> "PartialSymbolStack":
        return cls(symbols=(), variable=variable)

    def has_variable(self) -> bool:
        return self.variable is not None

    def can_match_empty(self) -> bool:
        return len(self.symbols) == 0

    def can_only_match_empty(self) -> bool:
        return len(self.symbols) == 0 and self.variable is None

    def contains_symbols(self) -> bool:
        return len(self.symbols) > 0

    def with_offset(self, symbol_offset: int, scope_offset: int) -> "PartialSymbolStack":
        if symbol_offset == 0 and scope_offset == 0:
            return self
        if self.variable is None:
            variable = None
        elif symbol_offset == 0:
            variable = self.variable
        else:
            variable = self.variable.with_offset(symbol_offset)
        symbols = tuple(symbol.with_offset(scope_offset) for symbol in self.symbols)
        return PartialSymbolStack(symbols=symbols, variable=variable)

    def largest_symbol_stack_var(self) -> int:
        if self.variable is None:
            return 0
        return self.variable.as_int()

    def largest_scope_stack_var(self) -> int:
        max_var = 0
        for symbol in self.symbols:
            max_var = max(max_var, symbol.largest_scope_stack_var())
        return max_var

    def apply_bindings(
        self,
        symbol_bindings: "PartialSymbolStackBindings",
        scope_bindings: "PartialScopeStackBindings",
    ) -> "PartialSymbolStack":
        bound_stack = None
        if self.variable is not None:
            bound_stack = symbol_bindings.get(self.variable)
        base_symbols: tuple[PartialScopedSymbol, ...] = ()
        base_variable = self.variable
        if bound_stack is not None:
            base_symbols = tuple(
                symbol.apply_bindings(scope_bindings) for symbol in bound_stack.symbols
            )
            base_variable = bound_stack.variable
        prefix_symbols = tuple(symbol.apply_bindings(scope_bindings) for symbol in self.symbols)
        return PartialSymbolStack(
            symbols=prefix_symbols + base_symbols,
            variable=base_variable,
        )

    def unify(
        self,
        rhs: "PartialSymbolStack",
        symbol_bindings: "PartialSymbolStackBindings",
        scope_bindings: "PartialScopeStackBindings",
    ) -> "PartialSymbolStack":
        common_len = min(len(self.symbols), len(rhs.symbols))
        prefix = tuple(
            self.symbols[index].unify(rhs.symbols[index], scope_bindings)
            for index in range(common_len)
        )

        lhs_suffix_symbols = self.symbols[common_len:]
        rhs_suffix_symbols = rhs.symbols[common_len:]

        if len(lhs_suffix_symbols) == 0 and len(rhs_suffix_symbols) == 0:
            return _unify_symbol_suffix_empty(prefix, self, rhs, symbol_bindings, scope_bindings)
        if len(rhs_suffix_symbols) == 0:
            bindings = _SymbolUnifyBindings(
                symbol_bindings=symbol_bindings,
                scope_bindings=scope_bindings,
            )
            return _unify_symbol_rhs_is_prefix_of_lhs(
                prefix, self, rhs, lhs_suffix_symbols, bindings
            )
        if len(lhs_suffix_symbols) == 0:
            bindings = _SymbolUnifyBindings(
                symbol_bindings=symbol_bindings,
                scope_bindings=scope_bindings,
            )
            return _unify_symbol_lhs_is_prefix_of_rhs(
                prefix, self, rhs, rhs_suffix_symbols, bindings
            )
        _raise_symbol_stack_unsatisfied()
        raise AssertionError  # unreachable



from blends.models import (
    NId,
)
from blends.query import (
    get_node_by_path,
    match_ast_d,
    match_ast_group,
    match_ast_group_d,
)
from blends.syntax.builders.import_global import (
    build_import_global_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def resolve_imports(
    args: SyntaxGraphArgs,
    expression: str,
    imports_n_ids: dict[str, list[NId]],
) -> tuple[str, str | None, set[NId]]:
    graph = args.ast_graph
    nodes = graph.nodes
    module_nodes: set[NId] = set()
    alias = None

    if (
        (namespace_n_ids := imports_n_ids.get("namespace_import"))
        and (identifier_n_id := match_ast_d(graph, namespace_n_ids[0], "identifier"))
    ) or (
        (identifier_n_ids := imports_n_ids.get("identifier"))
        and (identifier_n_id := identifier_n_ids[0])
    ):
        alias = nodes[identifier_n_id].get("label_text")

    elif (named_n_ids := imports_n_ids.get("named_imports")) and (
        specifier_n_ids := match_ast_group_d(graph, named_n_ids[0], "import_specifier")
    ):
        for specifier_n_id in specifier_n_ids:
            module_nodes.add(specifier_n_id)
            identifier_n_id = nodes[specifier_n_id].get("label_field_name")
            alias = None
    if imports_n_ids.get("identifier") and (named_n_ids := imports_n_ids.get("named_imports")):
        module_nodes.update(match_ast_group_d(graph, named_n_ids[0], "import_specifier"))
    return expression, alias, module_nodes


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes
    alias: str | None = None
    module_nodes: set[NId] = set()
    expression = ""

    if exp_n_id := nodes[args.n_id].get("label_field_source"):
        expression = node_to_str(graph, exp_n_id)[1:-1]
    import_clauses = match_ast_group(graph, args.n_id, "import_clause", "import_require_clause")
    clause_n_ids = import_clauses.get("import_clause") or import_clauses.get(
        "import_require_clause",
    )

    if (clause_n_ids) and (
        imports_n_ids := match_ast_group(
            graph,
            clause_n_ids[0],
            "namespace_import",
            "named_imports",
            "identifier",
        )
    ):
        expression, alias, module_nodes = resolve_imports(args, expression, imports_n_ids)
    if (import_req_n_id := import_clauses.get("import_require_clause")) and (
        exp_id := get_node_by_path(graph, import_req_n_id[0], "string", "string_fragment")
    ):
        expression = nodes[exp_id].get("label_text", "")

    return build_import_global_node(args, expression, module_nodes, alias)



from blends.models import (
    NId,
    SyntaxGraph,
)
from blends.query import (
    match_ast_group_d,
    pred_ast,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
)
from blends.symbolic_evaluation.utils import (
    get_lookup_path,
)


def is_not_valid_expression(
    args: SymbolicEvalArgs, graph: SyntaxGraph, n_attrs: dict[str, str]
) -> bool:
    return "." in n_attrs["expression"] or graph.nodes[pred_ast(graph, args.n_id)[0]][
        "label_type"
    ] not in {"Assignment", "VariableDeclaration"}


def solve(args: SymbolicEvalArgs) -> NId | None:
    graph = args.graph
    n_attrs = graph.nodes[args.n_id]
    if is_not_valid_expression(args, graph, n_attrs):
        return None

    try:
        search_path = get_lookup_path(graph, args.path, args.n_id)
    except ValueError:
        return None

    search_expr = n_attrs["expression"]
    search_member = n_attrs["member"]
    for n_id in search_path:
        if graph.nodes[n_id]["label_type"] == "Assignment" and search_in_assignment(
            graph,
            n_id,
            search_expr,
            search_member,
        ):
            return n_id
        if graph.nodes[n_id]["label_type"] == "VariableDeclaration" and (
            val_id := search_in_declaration(graph, n_id, search_expr, search_member)
        ):
            return val_id
    return None


def search_in_assignment(
    graph: SyntaxGraph, n_id: NId, search_expr: str, search_member: str
) -> bool:
    var_id = graph.nodes[n_id]["variable_id"]
    return (
        graph.nodes[var_id]["label_type"] == "ElementAccess"
        and (v_attrs := graph.nodes[var_id]) is not None
        and graph.nodes[v_attrs["expression_id"]].get("symbol") == search_expr
        and graph.nodes[v_attrs["arguments_id"]].get("value") == search_member
    )


def search_in_objects(graph: SyntaxGraph, val_id: NId, search_member: str) -> NId | None:
    childs = match_ast_group_d(graph, val_id, "Pair")
    if pair_id := next(
        (
            _id
            for _id in childs
            if graph.nodes[graph.nodes[_id]["key_id"]].get("symbol") == search_member
        ),
        None,
    ):
        return NId(graph.nodes[pair_id]["value_id"])
    return None


def search_in_declaration(
    graph: SyntaxGraph,
    n_id: NId,
    search_expr: str,
    search_member: str,
) -> NId | None:
    if (
        graph.nodes[n_id]["variable"] == search_expr
        and (val_id := graph.nodes[n_id].get("value_id"))
        and graph.nodes[val_id]["label_type"] == "Object"
    ):
        return search_in_objects(graph, val_id, search_member)
    return None



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    var_name = ""
    if identifiers := match_ast_group_d(graph, args.n_id, "identifier"):
        var_name = ".".join(
            graph.nodes[_id]["label_text"]
            for _id in identifiers
            if "label_text" in graph.nodes[_id]
        )
    return build_symbol_lookup_node(args, var_name)



from enum import Enum


class StackGraphNodeKind(str, Enum):
    DROP_SCOPES = "DropScopesNode"
    JUMP_TO = "JumpToNode"
    POP_SCOPED_SYMBOL = "PopScopedSymbolNode"
    POP_SYMBOL = "PopSymbolNode"
    PUSH_SCOPED_SYMBOL = "PushScopedSymbolNode"
    PUSH_SYMBOL = "PushSymbolNode"
    ROOT = "RootNode"
    SCOPE = "ScopeNode"


_VALID_STACK_GRAPH_KINDS = {kind.value for kind in StackGraphNodeKind}


def is_stack_graph_kind(value: object) -> bool:
    if not isinstance(value, str):
        return False
    return value in _VALID_STACK_GRAPH_KINDS



from blends.query import (
    adj_ast,
    lookup_first_cfg_parent,
)
from blends.symbolic_evaluation.context.search import (
    search_until_def,
)
from blends.symbolic_evaluation.dispatcher import (
    generic,
)
from blends.symbolic_evaluation.models import (
    SymbolicEvalArgs,
    SymbolicEvaluation,
)
from blends.symbolic_evaluation.utils import (
    get_lookup_path,
)

OUTSIDEPATH_TYPES = {
    "FieldDeclaration",
    "Import",
    "MethodDeclaration",
    "MethodInvocation",
    "Parameter",
    "VariableDeclaration",
}


def evaluate(args: SymbolicEvalArgs) -> SymbolicEvaluation:
    symbol_id = args.n_id
    args.evaluation[symbol_id] = False
    symbol = args.graph.nodes[args.n_id]["symbol"]

    try:
        path = get_lookup_path(args.graph, args.path, symbol_id)
    except ValueError:
        path = args.path

    refs_search_order = list(search_until_def(args.graph, path, symbol))
    refs_exec_order = reversed(refs_search_order)

    args.evaluation[symbol_id] = False
    arg_ids = adj_ast(args.graph, symbol_id)
    danger = [generic(args.fork_n_id(arg_id)).danger for arg_id in arg_ids]
    refs_dangers = []
    for ref_id in refs_exec_order:
        if args.graph.nodes[ref_id]["label_type"] in OUTSIDEPATH_TYPES:
            generic(args.fork_n_id(ref_id))
        elif ref_id in args.path:
            cfg_id = lookup_first_cfg_parent(args.graph, ref_id)
            generic(args.fork_n_id(cfg_id))

        if ref_id in args.evaluation:
            refs_dangers.append(args.evaluation[ref_id])

    args.evaluation[symbol_id] = any(refs_dangers) or any(danger)

    if args.method_evaluators and (method_evaluator := args.method_evaluators.get("symbol_lookup")):
        args.evaluation[args.n_id] = method_evaluator(args).danger
    return SymbolicEvaluation(args.evaluation[args.n_id], args.triggers)



from blends.models import (
    NId,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    name = "AnonymousMethod"

    block_id = graph.nodes[args.n_id].get("label_field_body")
    parameters_id = graph.nodes[args.n_id].get("label_field_parameters")

    direct: DirectChildren = {}
    if block_id:
        direct["block_id"] = block_id
    if parameters_id is not None:
        direct["parameters_id"] = parameters_id

    return build_method_declaration_node(args, name, direct, {})



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.for_statement import (
    build_for_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    body_id = n_attrs["label_field_body"]
    children = adj_ast(graph, args.n_id)
    var_node: NId | None = None
    condition_node: NId | None = None
    for child in children:
        if graph.nodes[child]["label_type"] == "range_clause":
            var_node = graph.nodes[child].get("label_field_left")
            condition_node = graph.nodes[child].get("label_field_right")

    return build_for_statement_node(args, var_node, condition_node, None, body_id)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.c_sharp import (
    accessor_declaration as cs_accessor_declaration,
)
from blends.syntax.readers.c_sharp import (
    anonymous_object_creation as cs_anon_object_creation,
)
from blends.syntax.readers.c_sharp import (
    argument as cs_argument,
)
from blends.syntax.readers.c_sharp import (
    argument_list as cs_argument_list,
)
from blends.syntax.readers.c_sharp import (
    array_creation_expression as cs_array_creation_expression,
)
from blends.syntax.readers.c_sharp import (
    arrow_expression_clause as cs_arrow_expression_clause,
)
from blends.syntax.readers.c_sharp import (
    assignment_expression as cs_assignment_expression,
)
from blends.syntax.readers.c_sharp import (
    attribute as cs_attribute,
)
from blends.syntax.readers.c_sharp import (
    attribute_list as cs_attribute_list,
)
from blends.syntax.readers.c_sharp import (
    binary_expression as cs_binary_expression,
)
from blends.syntax.readers.c_sharp import (
    boolean_literal as cs_boolean_literal,
)
from blends.syntax.readers.c_sharp import (
    bracketed_argument_list as cs_bracketed_argument_list,
)
from blends.syntax.readers.c_sharp import (
    break_statement as cs_break_statement,
)
from blends.syntax.readers.c_sharp import (
    cast_expression as cs_cast_expression,
)
from blends.syntax.readers.c_sharp import (
    catch_clause as cs_catch_clause,
)
from blends.syntax.readers.c_sharp import (
    catch_declaration as cs_catch_declaration,
)
from blends.syntax.readers.c_sharp import (
    class_declaration as cs_class_declaration,
)
from blends.syntax.readers.c_sharp import (
    comment as cs_comment,
)
from blends.syntax.readers.c_sharp import (
    compilation_unit as cs_compilation_unit,
)
from blends.syntax.readers.c_sharp import (
    conditional_access_expression as cs_conditional_access_expression,
)
from blends.syntax.readers.c_sharp import (
    constructor_declaration as cs_constructor_declaration,
)
from blends.syntax.readers.c_sharp import (
    continue_statement as cs_continue_statement,
)
from blends.syntax.readers.c_sharp import (
    declaration_block as cs_declaration_block,
)
from blends.syntax.readers.c_sharp import (
    do_statement as cs_do_statement,
)
from blends.syntax.readers.c_sharp import (
    element_access_expression as cs_element_access_expression,
)
from blends.syntax.readers.c_sharp import (
    element_binding_expression as cs_element_binding_expression,
)
from blends.syntax.readers.c_sharp import (
    execution_block as cs_execution_block,
)
from blends.syntax.readers.c_sharp import (
    expression_statement as cs_expression_statement,
)
from blends.syntax.readers.c_sharp import (
    file_scoped_namespace_declaration as cs_file_scoped_namespace_decla,
)
from blends.syntax.readers.c_sharp import (
    finally_clause as cs_finally_clause,
)
from blends.syntax.readers.c_sharp import (
    for_each_statement as cs_for_each_statement,
)
from blends.syntax.readers.c_sharp import (
    for_statement as cs_for_statement,
)
from blends.syntax.readers.c_sharp import (
    global_statement as cs_global_statement,
)
from blends.syntax.readers.c_sharp import (
    identifier as cs_identifier,
)
from blends.syntax.readers.c_sharp import (
    if_statement as cs_if_statement,
)
from blends.syntax.readers.c_sharp import (
    implicit_parameter as cs_implicit_parameter,
)
from blends.syntax.readers.c_sharp import (
    initializer_expression as cs_initializer_expression,
)
from blends.syntax.readers.c_sharp import (
    interface_declaration as cs_interface_declaration,
)
from blends.syntax.readers.c_sharp import (
    interpolated_string_expression as cs_interpolated_string_expression,
)
from blends.syntax.readers.c_sharp import (
    interpolation as cs_interpolation,
)
from blends.syntax.readers.c_sharp import (
    invocation_expression as cs_invocation_expression,
)
from blends.syntax.readers.c_sharp import (
    lambda_expression as cs_lambda_expression,
)
from blends.syntax.readers.c_sharp import (
    local_declaration_statement as cs_local_declaration_statement,
)
from blends.syntax.readers.c_sharp import (
    member_access_expression as cs_member_access_expression,
)
from blends.syntax.readers.c_sharp import (
    member_binding_expression as cs_member_binding_expression,
)
from blends.syntax.readers.c_sharp import (
    method_declaration as cs_method_declaration,
)
from blends.syntax.readers.c_sharp import (
    name_equals as cs_name_equals,
)
from blends.syntax.readers.c_sharp import (
    namespace_declaration as cs_namespace_declaration,
)
from blends.syntax.readers.c_sharp import (
    null_literal as cs_null_literal,
)
from blends.syntax.readers.c_sharp import (
    number_literal as cs_number_literal,
)
from blends.syntax.readers.c_sharp import (
    object_creation_expression as cs_object_creation_expression,
)
from blends.syntax.readers.c_sharp import (
    parameter as cs_parameter,
)
from blends.syntax.readers.c_sharp import (
    parameter_list as cs_parameter_list,
)
from blends.syntax.readers.c_sharp import (
    parenthesized_expression as cs_parenthesized_expression,
)
from blends.syntax.readers.c_sharp import (
    postfix_unary_expression as cs_postfix_unary_expression,
)
from blends.syntax.readers.c_sharp import (
    prefix_expression as cs_prefix_expression,
)
from blends.syntax.readers.c_sharp import (
    property_declaration as cs_property_declaration,
)
from blends.syntax.readers.c_sharp import (
    return_statement as cs_return_statement,
)
from blends.syntax.readers.c_sharp import (
    string_literal as cs_string_literal,
)
from blends.syntax.readers.c_sharp import (
    switch_body as cs_switch_body,
)
from blends.syntax.readers.c_sharp import (
    switch_section as cs_switch_section,
)
from blends.syntax.readers.c_sharp import (
    switch_statement as cs_switch_statement,
)
from blends.syntax.readers.c_sharp import (
    this_expression as cs_this_expression,
)
from blends.syntax.readers.c_sharp import (
    this_node as cs_this_node,
)
from blends.syntax.readers.c_sharp import (
    throw_statement as cs_throw_statement,
)
from blends.syntax.readers.c_sharp import (
    try_statement as cs_try_statement,
)
from blends.syntax.readers.c_sharp import (
    type_of_expression as cs_type_of_expression,
)
from blends.syntax.readers.c_sharp import (
    type_parameter_list as cs_type_parameter_list,
)
from blends.syntax.readers.c_sharp import (
    unary_expression as cs_unary_expression,
)
from blends.syntax.readers.c_sharp import (
    using_global_directive as cs_using_global_directive,
)
from blends.syntax.readers.c_sharp import (
    using_statement as cs_using_statement,
)
from blends.syntax.readers.c_sharp import (
    variable_declaration as cs_variable_declaration,
)
from blends.syntax.readers.c_sharp import (
    while_statement as cs_while_statement,
)

C_SHARP_DISPATCHERS: Dispatcher = {
    "accessor_declaration": cs_accessor_declaration.reader,
    "anonymous_object_creation_expression": cs_anon_object_creation.reader,
    "argument": cs_argument.reader,
    "attribute_argument": cs_argument.reader,
    "argument_list": cs_argument_list.reader,
    "attribute_argument_list": cs_argument_list.reader,
    "array_creation_expression": cs_array_creation_expression.reader,
    "implicit_array_creation_expression": cs_array_creation_expression.reader,
    "arrow_expression_clause": cs_arrow_expression_clause.reader,
    "assignment_expression": cs_assignment_expression.reader,
    "attribute": cs_attribute.reader,
    "attribute_list": cs_attribute_list.reader,
    "binary_expression": cs_binary_expression.reader,
    "boolean_literal": cs_boolean_literal.reader,
    "bracketed_argument_list": cs_bracketed_argument_list.reader,
    "break_statement": cs_break_statement.reader,
    "block": cs_execution_block.reader,
    "cast_expression": cs_cast_expression.reader,
    "catch_clause": cs_catch_clause.reader,
    "catch_declaration": cs_catch_declaration.reader,
    "class_declaration": cs_class_declaration.reader,
    "comment": cs_comment.reader,
    "compilation_unit": cs_compilation_unit.reader,
    "conditional_access_expression": cs_conditional_access_expression.reader,
    "constructor_declaration": cs_constructor_declaration.reader,
    "continue_statement": cs_continue_statement.reader,
    "declaration_list": cs_declaration_block.reader,
    "do_statement": cs_do_statement.reader,
    "element_access_expression": cs_element_access_expression.reader,
    "element_binding_expression": cs_element_binding_expression.reader,
    "expression_statement": cs_expression_statement.reader,
    "field_declaration": cs_variable_declaration.reader,
    "file_scoped_namespace_declaration": cs_file_scoped_namespace_decla.reader,
    "finally_clause": cs_finally_clause.reader,
    "for_each_statement": cs_for_each_statement.reader,
    "for_statement": cs_for_statement.reader,
    "global_statement": cs_global_statement.reader,
    "identifier": cs_identifier.reader,
    "conditional_expression": cs_if_statement.reader,
    "if_statement": cs_if_statement.reader,
    "initializer_expression": cs_initializer_expression.reader,
    "interface_declaration": cs_interface_declaration.reader,
    "interpolated_string_expression": cs_interpolated_string_expression.reader,
    "invocation_expression": cs_invocation_expression.reader,
    "interpolation": cs_interpolation.reader,
    "interpolation_brace": cs_interpolated_string_expression.reader,
    "lambda_expression": cs_lambda_expression.reader,
    "local_declaration_statement": cs_local_declaration_statement.reader,
    "member_access_expression": cs_member_access_expression.reader,
    "member_binding_expression": cs_member_binding_expression.reader,
    "method_declaration": cs_method_declaration.reader,
    "local_function_statement": cs_method_declaration.reader,
    "name_equals": cs_name_equals.reader,
    "namespace_declaration": cs_namespace_declaration.reader,
    "null_literal": cs_null_literal.reader,
    "integer_literal": cs_number_literal.reader,
    "real_literal": cs_number_literal.reader,
    "object_creation_expression": cs_object_creation_expression.reader,
    "parameter": cs_parameter.reader,
    "implicit_parameter": cs_implicit_parameter.reader,
    "parameter_list": cs_parameter_list.reader,
    "implicit_parameter_list": cs_parameter_list.reader,
    "parenthesized_expression": cs_parenthesized_expression.reader,
    "postfix_unary_expression": cs_postfix_unary_expression.reader,
    "prefix_unary_expression": cs_prefix_expression.reader,
    "property_declaration": cs_property_declaration.reader,
    "return_statement": cs_return_statement.reader,
    "character_literal": cs_string_literal.reader,
    "string_literal": cs_string_literal.reader,
    "verbatim_string_literal": cs_string_literal.reader,
    "predefined_type": cs_string_literal.reader,
    "switch_body": cs_switch_body.reader,
    "switch_statement": cs_switch_statement.reader,
    "switch_section": cs_switch_section.reader,
    "this_expression": cs_this_expression.reader,
    "this": cs_this_node.reader,
    "throw_statement": cs_throw_statement.reader,
    "type_of_expression": cs_type_of_expression.reader,
    "type_parameter_list": cs_type_parameter_list.reader,
    "try_statement": cs_try_statement.reader,
    "using_directive": cs_using_global_directive.reader,
    "using_statement": cs_using_statement.reader,
    "unary_expression": cs_unary_expression.reader,
    "variable_declaration": cs_variable_declaration.reader,
    "while_statement": cs_while_statement.reader,
}
